import collections
import glob
import logging
import logging.config as logging_config
import os
from inspect import stack

import numpy as np
import six
from six import string_types

from .__version__ import VERSION, VERSION_STATUS, __version__


from mdt.configuration import get_logging_configuration_dict
try:
    logging_config.dictConfig(get_logging_configuration_dict())
except ValueError:
    print('Logging disabled')


from mdt.user_script_info import easy_save_user_script_info
from mdt.utils import estimate_noise_std, get_cl_devices, load_problem_data, create_blank_mask, create_index_matrix, \
    volume_index_to_roi_index, roi_index_to_volume_index, load_brain_mask, init_user_settings, restore_volumes, \
    apply_mask, create_roi, volume_merge, concatenate_mri_sets, create_median_otsu_brain_mask, load_samples, \
    load_nifti, write_slice_roi, split_write_dataset, apply_mask_to_file, extract_volumes, recalculate_error_measures, \
    create_signal_estimates, get_slice_in_dimension, per_model_logging_context, get_temporary_results_dir
from mdt.batch_utils import collect_batch_fit_output, run_function_on_batch_fit_output
from mdt.protocols import load_bvec_bval, load_protocol, auto_load_protocol, write_protocol, write_bvec_bval
from mdt.components_loader import load_component, get_model
from mdt.configuration import config_context, get_processing_strategy
from mdt.exceptions import InsufficientProtocolError


__author__ = 'Robbert Harms'
__date__ = "2015-03-10"
__license__ = "LGPL v3"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


def fit_model(model, problem_data, output_folder, optimizer=None,
              recalculate=False, only_recalculate_last=False, cascade_subdir=False,
              cl_device_ind=None, double_precision=False, tmp_results_dir=True, save_user_script_info=True):
    """Run the optimizer on the given model.

    Args:
        model (str or :class:`~mdt.models.composite.DMRICompositeModel` or :class:`~mdt.models.cascade.DMRICascadeModelInterface`):
            An implementation of an AbstractModel that contains the model we want to optimize or the name of
            an model.
        problem_data (:class:`~mdt.utils.DMRIProblemData`): the problem data object containing all the info needed for
            diffusion MRI model fitting
        output_folder (string): The path to the folder where to place the output, we will make a subdir with the
            model name in it.
        optimizer (:class:`mot.cl_routines.optimizing.base.AbstractOptimizer`): The optimization routine to use.
        recalculate (boolean): If we want to recalculate the results if they are already present.
        only_recalculate_last (boolean):
            This is only of importance when dealing with CascadeModels.
            If set to true we only recalculate the last element in the chain (if recalculate is set to True, that is).
            If set to false, we recalculate everything. This only holds for the first level of the cascade.
        cascade_subdir (boolean): if we want to create a subdirectory for the given model if it is a cascade model.
            Per default we output the maps of cascaded results in the same directory, this allows reusing cascaded
            results for other cascades (for example, if you cascade BallStick -> Noddi you can use the BallStick results
            also for BallStick -> Charmed). This flag disables that behaviour and instead outputs the results of
            a cascade model to a subdirectory for that cascade. This does not apply recursive.
        cl_device_ind (int or list): the index of the CL device to use. The index is from the list from the function
            utils.get_cl_devices(). This can also be a list of device indices.
        double_precision (boolean): if we would like to do the calculations in double precision
        tmp_results_dir (str, True or None): The temporary dir for the calculations. Set to a string to use
            that path directly, set to True to use the config value, set to None to disable.
        save_user_script_info (boolean, str or SaveUserScriptInfo): The info we need to save about the script the
            user is currently executing. If True (default) we use the stack to lookup the script the user is executing
            and save that using a SaveFromScript saver. If a string is given we use that filename again for the
            SaveFromScript saver. If False or None, we do not write any information. If a SaveUserScriptInfo is
            given we use that directly.

    Returns:
        dict: The result maps for the (final) optimized model.
                This returns the results as 2d arrays with on the first dimension the optimized voxels
                and on the second the value(s) for the micro-structure maps.
    """
    import mdt.utils
    from mdt.model_fitting import ModelFit

    if not mdt.utils.check_user_components():
        init_user_settings(pass_if_exists=True)

    model_fit = ModelFit(model, problem_data, output_folder, optimizer=optimizer, recalculate=recalculate,
                         only_recalculate_last=only_recalculate_last,
                         cascade_subdir=cascade_subdir,
                         cl_device_ind=cl_device_ind, double_precision=double_precision,
                         tmp_results_dir=tmp_results_dir)

    results = model_fit.run()
    easy_save_user_script_info(save_user_script_info, output_folder + '/used_scripts.py',
                               stack()[1][0].f_globals.get('__file__'))
    return results


def sample_model(model, problem_data, output_folder, sampler=None, recalculate=False,
                 cl_device_ind=None, double_precision=False, store_samples=True, tmp_results_dir=True,
                 save_user_script_info=True, initialization_maps=None):
    """Sample a composite model using the given cascading strategy.

    Args:
        model (:class:`~mdt.models.composite.DMRICompositeModel` or str): the model to sample
        problem_data (:class:`~mdt.utils.DMRIProblemData`): the problem data object
        output_folder (string): The path to the folder where to place the output, we will make a subdir with the
            model name in it (for the optimization results) and then a subdir with the samples output.
        sampler (:class:`mot.cl_routines.sampling.base.AbstractSampler`): the sampler to use
        recalculate (boolean): If we want to recalculate the results if they are already present.
        cl_device_ind (int): the index of the CL device to use. The index is from the list from the function
            utils.get_cl_devices().
        double_precision (boolean): if we would like to do the calculations in double precision
        store_samples (boolean): if set to False we will store none of the samples. Use this
                if you are only interested in the volume maps and not in the entire sample chain.
        tmp_results_dir (str, True or None): The temporary dir for the calculations. Set to a string to use
                that path directly, set to True to use the config value, set to None to disable.
        save_user_script_info (boolean, str or SaveUserScriptInfo): The info we need to save about the script the
            user is currently executing. If True (default) we use the stack to lookup the script the user is executing
            and save that using a SaveFromScript saver. If a string is given we use that filename again for the
            SaveFromScript saver. If False or None, we do not write any information. If a SaveUserScriptInfo is
            given we use that directly.
        initialization_maps (dict): 4d maps to initialize the sampling with. Per default this is None,
            common practice is to use the maps from an optimization as starting point

    Returns:
        dict: the samples per parameter as a numpy memmap if store_samples is True
    """
    import mdt.utils
    from mot.load_balance_strategies import EvenDistribution
    from mdt.model_sampling import sample_composite_model
    from mdt.models.cascade import DMRICascadeModelInterface
    import mot.configuration

    if not mdt.utils.check_user_components():
        init_user_settings(pass_if_exists=True)

    if isinstance(model, string_types):
        model = get_model(model)

    if isinstance(model, DMRICascadeModelInterface):
        raise ValueError('The function \'sample_model()\' does not accept cascade models.')

    if not model.is_protocol_sufficient(problem_data.protocol):
        raise InsufficientProtocolError(
            'The given protocol is insufficient for this model. '
            'The reported errors where: {}'.format(model.get_protocol_problems(problem_data.protocol)))

    if cl_device_ind is not None and not isinstance(cl_device_ind, collections.Iterable):
        cl_device_ind = [cl_device_ind]

    if cl_device_ind is None:
        cl_context_action = mot.configuration.VoidConfigurationAction()
    else:
        cl_context_action = mot.configuration.RuntimeConfigurationAction(
            cl_environments=[get_cl_devices()[ind] for ind in cl_device_ind],
            load_balancer=EvenDistribution())

    with mot.configuration.config_context(cl_context_action):
        if sampler is None:
            sampler = configuration.get_sampler()

        processing_strategy = get_processing_strategy('sampling', model_names=model.name)
        processing_strategy.set_tmp_dir(get_temporary_results_dir(tmp_results_dir))

        output_folder = os.path.join(output_folder, model.name, 'samples')
        if not os.path.isdir(output_folder):
            os.makedirs(output_folder)

        with per_model_logging_context(output_folder, overwrite=recalculate):
            logger = logging.getLogger(__name__)
            logger.info('Using MDT version {}'.format(__version__))
            logger.info('Preparing for model {0}'.format(model.name))

            if initialization_maps:
                model.set_initial_parameters(create_roi(initialization_maps, problem_data.mask))

            model.double_precision = double_precision

            results = sample_composite_model(model, problem_data, output_folder, sampler,
                                          processing_strategy, recalculate=recalculate, store_samples=store_samples)

        easy_save_user_script_info(save_user_script_info, output_folder + '/used_scripts.py',
                                   stack()[1][0].f_globals.get('__file__'))
        return results


def batch_fit(data_folder, batch_profile=None, subjects_selection=None, recalculate=False,
              models_to_fit=None, cascade_subdir=False, cl_device_ind=None, dry_run=False,
              double_precision=False, tmp_results_dir=True):
    """Run all the available and applicable models on the data in the given folder.

    Args:
        data_folder (str): The data folder to process
        batch_profile (:class:`~mdt.batch_utils.BatchProfile` or str): the batch profile to use,
            or the name of a batch profile to use. If not given it is auto detected.
        subjects_selection (:class:`~mdt.batch_utils.BatchSubjectSelection`): the subjects to use for processing.
            If None all subjects are processed.
        recalculate (boolean): If we want to recalculate the results if they are already present.
        models_to_fit (list of str): A list of models to fit to the data. This overrides the models in
                the batch config.
        cascade_subdir (boolean): if we want to create a subdirectory for every cascade model.
            Per default we output the maps of cascaded results in the same directory, this allows reusing cascaded
            results for other cascades (for example, if you cascade BallStick -> Noddi you can use the BallStick results
            also for BallStick -> Charmed). This flag disables that behaviour and instead outputs the results of
            a cascade model to a subdirectory for that cascade. This does not apply recursive.
        cl_device_ind (int or list of int): the index of the CL device to use.
            The index is from the list from the function get_cl_devices().
        dry_run (boolean): a dry run will do no computations, but will list all the subjects found in the
            given directory.
        double_precision (boolean): if we would like to do the calculations in double precision
        tmp_results_dir (str, True or None): The temporary dir for the calculations. Set to a string to use
                that path directly, set to True to use the config value, set to None to disable.

    Returns:
        The list of subjects we will calculate / have calculated.
    """
    import mdt.utils
    from mdt.model_fitting import BatchFitting

    if not mdt.utils.check_user_components():
        init_user_settings(pass_if_exists=True)

    batch_fitting = BatchFitting(data_folder, batch_profile=batch_profile, subjects_selection=subjects_selection,
                                 recalculate=recalculate, models_to_fit=models_to_fit, cascade_subdir=cascade_subdir,
                                 cl_device_ind=cl_device_ind, double_precision=double_precision,
                                 tmp_results_dir=tmp_results_dir)

    if dry_run:
        return batch_fitting.get_subjects_info()

    return batch_fitting.run()


def view_maps(data, config=None, to_file=None, to_file_options=None,
              block=True, show_maximized=False, use_qt=True, figure_options=None,
              window_title=None, enable_directory_watcher=True):
    """View a number of maps using the MDT Maps Visualizer.

    To save a file to an image, you can use (the simplest) the following two options::

        to_file='filename.png',
        figure_options={'width': 1200, 'height': 750, 'dpi': 100}

    Args:
        data (str, dict, :class:`~mdt.visualization.maps.base.DataInfo`): the data we are showing,
            either a dictionary with result maps, a string with a path name or a DataInfo object
        config (str, dict, :class:`~mdt.visualization.maps.base import MapPlotConfig`): either a Yaml string or a
            dictionary with configuration settings or a ValidatedMapPlotConfig object to use directly
        to_file (str): if set we output the figure to a file and do not launch a GUI
        to_file_options (dict): extra output options for the savefig command from matplotlib, if dpi is not given, we
            use the dpi from the figure_options.
        block (boolean): if we block the plots or not
        show_maximized (boolean): if we show the window maximized or not
        use_qt (boolean): if we want to use the Qt GUI, or show the results directly in matplotlib
        figure_options (dict): figure options for the matplotlib Figure, if figsizes is not given you can also specify
            two ints, width and height, to indicate the pixel size of the resulting figure, together with the dpi they
            are used to calculate the figsize.
        window_title (str): the title for the window
        enable_directory_watcher (boolean): if the directory watcher should be enabled/disabled, only applicable for the
            QT GUI. If the directory watcher is enabled, the viewer will automatically add new maps when added
            to the folder and also automatically remove maps when they are removed from the directory.
            It is useful to disable this if you want to have multiple viewers open with old results.
    """
    from mdt.gui.maps_visualizer.main import start_gui
    from mdt.visualization.maps.base import MapPlotConfig
    from mdt.visualization.maps.matplotlib_renderer import MapsVisualizer
    import matplotlib.pyplot as plt
    from mdt.visualization.maps.base import DataInfo

    if isinstance(data, string_types):
        data = DataInfo.from_dir(data)
    elif isinstance(data, dict):
        data = DataInfo(data)
    elif data is None:
        data = DataInfo({})

    if config is None:
        config = MapPlotConfig()
    elif isinstance(config, string_types):
        if config.strip():
            config = MapPlotConfig.from_yaml(config)
        else:
            config = MapPlotConfig()
    elif isinstance(config, dict):
        config = MapPlotConfig.from_dict(config)

    if to_file:
        figure_options = figure_options or {}

        figure_options['dpi'] = figure_options.get('dpi', 80)
        if 'figsize' not in figure_options:
            figure_options['figsize'] = (figure_options.pop('width', 800) / figure_options['dpi'],
                                         figure_options.pop('height', 640) / figure_options['dpi'])

        figure = plt.figure(**figure_options)
        viz = MapsVisualizer(data, figure)

        to_file_options = to_file_options or {}
        to_file_options['dpi'] = to_file_options.get('dpi', figure_options['dpi'])
        viz.to_file(to_file, config, **to_file_options)
    elif use_qt:
        start_gui(data, config, app_exec=block, show_maximized=show_maximized, window_title=window_title,
                  enable_directory_watcher=enable_directory_watcher)
    else:
        figure_options = figure_options or {}
        figure_options['dpi'] = figure_options.get('dpi', 100)
        if 'figsize' not in figure_options:
            figure_options['figsize'] = (figure_options.pop('width', 1800) / figure_options['dpi'],
                                         figure_options.pop('height', 1600) / figure_options['dpi'])

        figure = plt.figure(**figure_options)
        viz = MapsVisualizer(data, figure)
        viz.show(config, block=block, maximize=show_maximized)


def results_preselection_names(data):
    """Generate a list of useful map names to display.

    This is primarily to be used as argument to the config option ``maps_to_show`` in the function :func:`view_maps`.

    Args:
        data (str or dict or list of str): either a directory or a dictionary of results or a list of map names.

    Returns:
        list of str: the list of useful/filtered map names.
    """
    keys = []
    if isinstance(data, string_types):
        for extension in ('.nii', '.nii.gz'):
            for f in glob.glob(os.path.join(data, '*' + extension)):
                keys.append(os.path.basename(f)[0:-len(extension)])
    elif isinstance(data, dict):
        keys = data.keys()
    else:
        keys = data

    filter_match = ('.vec', '.d', '.sigma', 'AIC', 'Errors.mse', 'Errors.sse', '.eigen_ranking',
                    'SignalEstimates', 'UsedMask')
    return list(sorted(filter(lambda v: all(m not in v for m in filter_match), keys)))


def block_plots(use_qt=True):
    """A small function to block matplotlib plots and Qt GUI instances.

    This basically calls either ``plt.show()`` and ``QtApplication.exec_()`` depending on ``use_qt``.

    Args:
        use_qt (boolean): if True we block Qt windows, if False we block matplotlib windows
    """
    if use_qt:
        from mdt.gui.utils import QtManager
        QtManager.exec_()
    else:
        import matplotlib.pyplot as plt
        plt.show()


def view_result_samples(data, **kwargs):
    """View the samples from the given results set.

    Args:
        data (string or dict): The location of the maps to use the samples from, or the samples themselves.
        kwargs (dict): see SampleVisualizer for all the supported keywords
    """
    from mdt.visualization.samples import SampleVisualizer

    if isinstance(data, string_types):
        data = load_samples(data)

    if kwargs.get('voxel_ind') is None:
        kwargs.update({'voxel_ind': data[list(data.keys())[0]].shape[0] / 2})
    SampleVisualizer(data).show(**kwargs)


def make_path_joiner(*folder):
    """Generates and returns an instance of utils.PathJoiner to quickly join pathnames.

    Returns:
        PathJoiner: easy path manipulation path joiner
    """
    from mdt.utils import PathJoiner
    return PathJoiner(*folder)


def write_image(fname, data, header):
    """Write the given data with the given header to the given file.

    Args:
        fname (str): The output filename
        data (ndarray): The data to write
        header (nibabel header): The header to use
    """
    import nibabel as nib
    nib.Nifti1Image(data, None, header).to_filename(fname)


def auto_convert_to_trackmark(input_folder, model_name=None, output_folder=None):
    """Convert the nifti files in the given folder to Trackmark.

    This automatically loads the correct files based on the model name. This is normally the dirname of the given
    path. If that is not the case you can give the model name explicitly.

    By default it outputs the results to a folder named "trackmark" in the given input folder. This can of course
    be overridden using the output_folder parameter.

    Args:
        input_folder (str): the name of the input folder
        model_name (str): the name of the model, if not given we use the last dirname of the given path
        output_folder (str): the output folder, if not given we will output to a subfolder "trackmark" in the
            given directory.
    """
    from mdt.nifti import TrackMark
    TrackMark.auto_convert(input_folder, model_name=model_name, output_folder=output_folder)


def write_trackmark_rawmaps(data, output_folder, maps_to_convert=None):
    """Convert the given nifti files in the input folder to rawmaps in the output folder.

    Args:
        data (str or dict): the name of the input folder, of a dictionary with maps to save.
        output_folder (str): the name of the output folder. Defaults to <input_folder>/trackmark.
        maps_to_convert (:class:`list`): the list with the names of the maps we want to convert (without the extension).
    """
    from mdt.nifti import TrackMark

    if isinstance(data, six.string_types):
        volumes = load_volume_maps(data, map_names=maps_to_convert)
    else:
        volumes = data
        if maps_to_convert:
            volumes = {k: v for k, v in volumes.items() if k in maps_to_convert}
    TrackMark.write_rawmaps(output_folder, volumes)


def write_trackmark_tvl(output_tvl, vector_directions, vector_magnitudes, tvl_header=(1, 1.8, 0, 0)):
    """Write a list of vector directions with corresponding magnitude to a trackmark TVL file.

    Note that the length of the vector_directions and vector_magnitudes should correspond to each other. Next, we only
    use the first three elements in both lists.

    Args:
        output_tvl (str): the name of the output tvl
        vector_directions (list of str/ndarray): a list of 4d volumes with per voxel the normalized vector direction
        vector_magnitudes (list of str/ndarray): a list of 4d volumes with per voxel the vector magnitude.
        tvl_header (list or tuple): The list with header arguments for writing the TVL. See IO.TrackMark for specifics.
    """
    from mdt.nifti import TrackMark
    if len(vector_directions) != len(vector_magnitudes):
        raise ValueError('The length of the list of vector directions does not '
                         'match with the length of the list of vector magnitudes.')
    TrackMark.write_tvl_direction_pairs(output_tvl, tvl_header, list(zip(vector_directions, vector_magnitudes))[:3])


def sort_maps(maps_to_sort_on, extra_maps_to_sort=None, reversed_sort=False, sort_index_map=None):
    """Sort the given maps on the maps to sort on.

    This first creates a sort matrix to index the maps in sorted order per voxel. Next, it creates the output
    maps for the maps we sort on. If extra_maps_to_sort is given it should be of the same length as the maps_to_sort_on.

    Args:
        maps_to_sort_on (:class:`list`): a list of string (filenames) or ndarrays we will use and compare
        extra_maps_to_sort (:class:`list`) an additional list we will sort based on the indices in maps_to_sort. This should
            be of the same length as maps_to_sort_on.
        reversed_sort (boolean): if we want to sort from large to small instead of small to large.
        sort_index_map (ndarray): if given we use this sort index map instead of generating one by sorting the
            maps_to_sort_on.

    Returns:
        tuple: the first element is the list of sorted volumes, the second the list of extra sorted maps and the
            last is the sort index map used.
    """
    def load_maps(map_list):
        tmp = []
        for data in map_list:
            if isinstance(data, string_types):
                tmp.append(load_nifti(data).get_data())
            else:
                tmp.append(data)
        return tmp

    maps_to_sort_on = load_maps(maps_to_sort_on)
    if extra_maps_to_sort:
        extra_maps_to_sort = load_maps(extra_maps_to_sort)

        if len(extra_maps_to_sort) != len(maps_to_sort_on):
            raise ValueError('The length of the maps to sort on and the extra maps to sort do not match.')

    from mdt.utils import create_sort_matrix, sort_volumes_per_voxel

    if sort_index_map is None:
        sort_index_map = create_sort_matrix(np.concatenate([m for m in maps_to_sort_on], axis=3),
                                            reversed_sort=reversed_sort)
    elif isinstance(sort_index_map, string_types):
        sort_index_map = np.round(load_nifti(sort_index_map).get_data()).astype(np.int64)

    sorted_maps = sort_volumes_per_voxel(maps_to_sort_on, sort_index_map)
    if extra_maps_to_sort:
        sorted_extra_maps = sort_volumes_per_voxel(extra_maps_to_sort, sort_index_map)
        return sorted_maps, sorted_extra_maps, sort_index_map
    return sorted_maps, [], sort_index_map


def load_volume_maps(directory, map_names=None, deferred=True):
    """Read a number of Nifti volume maps from a directory.

    Args:
        directory (str): the directory from which we want to read a number of maps
        map_names (list or tuple): the names of the maps we want to use. If given we only use and return these maps.
        deferred (boolean): if True we return an deferred loading dictionary instead of a dictionary with the values
            loaded as arrays.

    Returns:
        dict: A dictionary with the volumes. The keys of the dictionary are the filenames (without the extension) of the
            files in the given directory.
    """
    from mdt.nifti import get_all_image_data
    return get_all_image_data(directory, map_names=map_names, deferred=deferred)


def get_volume_names(directory):
    """Get the names of the Nifti volume maps in the given directory.

    Args:
        directory: the directory to get the names of the available maps from.

    Returns:
        :class:`list`: A list with the names of the volumes.
    """
    from mdt.nifti import yield_nifti_info
    return list(sorted(el[1] for el in yield_nifti_info(directory)))


def write_volume_maps(maps, directory, header, overwrite_volumes=True):
    """Write a dictionary with maps to the given directory using the given header.

    Args:
        maps (dict): The maps with as keys the map names and as values 3d or 4d maps
        directory (str): The dir to write to
        header: The Nibabel Image Header
        overwrite_volumes (boolean): If we want to overwrite the volumes if they are present.
    """
    from mdt.nifti import write_all_as_nifti
    write_all_as_nifti(maps, directory, header, overwrite_volumes=overwrite_volumes)


def get_list_of_composite_models():
    """Get a list of all available composite models

    Returns:
        list of str: A list of all available composite model names.
    """
    from mdt.components_loader import CompositeModelsLoader
    return CompositeModelsLoader().list_all()


def get_list_of_cascade_models():
    """Get a list of all available cascade models

    Returns:
        list of str: A list of available cascade models
    """
    from mdt.components_loader import CascadeModelsLoader
    return CascadeModelsLoader().list_all()


def get_models_list():
    """Get a list of all available models, composite and cascade.

    Returns:
        list of str: A list of available model names.
    """
    l = get_list_of_cascade_models()
    l.extend(get_list_of_composite_models())
    return list(sorted(l))


def get_models_meta_info():
    """Get the meta information tags for all the models returned by get_models_list()

    Returns:
        dict of dict: The first dictionary indexes the model names to the meta tags, the second holds the meta
            information.
    """
    from mdt.components_loader import CompositeModelsLoader, CascadeModelsLoader
    sml = CompositeModelsLoader()
    cml = CascadeModelsLoader()

    meta_info = {}
    for model_loader in (sml, cml):
        models = model_loader.list_all()
        for model in models:
            meta_info.update({model: model_loader.get_meta_info(model)})
    return meta_info


def get_batch_profile(batch_profile_name, *args, **kwargs):
    """Load one of the batch profiles.

    This is short for load_component('batch_profiles', batch_profile_name).

    Args:
        batch_profile_name (str): the name of the batch profile to use

    Returns:
        BatchProfile: the batch profile for use in batch fitting routines.
    """
    return load_component('batch_profiles', batch_profile_name, *args, **kwargs)


def gui(base_dir=None, app_exec=True):
    """Start the model fitting GUI.

    Args:
        base_dir (str): the starting directory for the file opening actions
        app_exec (boolean): if true we execute the Qt application, set to false to disable.
            This is only important if you want to start this GUI from within an existing Qt application. If you
            leave this at true in that case, this will try to start a new Qt application which may create problems.
    """
    from mdt.gui.model_fit.qt_main import start_gui
    return start_gui(base_dir=base_dir, app_exec=app_exec)



"""The general components loader.

This modules consists of two main items, component sources and component loaders. The component loaders have a list
of sources from which they load the available components.
"""
import inspect
import os
import imp #todo in P3.4 replace imp calls with importlib.SourceFileLoader(name, path).load_module(name)

import collections
from six import with_metaclass

from mdt.exceptions import NonUniqueComponent
from mot.model_building.cl_functions.base import ModelFunction, LibraryFunction
import mot.model_building.cl_functions.library_functions
import mot.model_building.cl_functions.model_functions

__author__ = 'Robbert Harms'
__date__ = "2015-06-21"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


def get_model(model_name, **kwargs):
    """Load one of the available models.

    Args:
        model_name (str): One of the models from get_list_of_composite_models() or get_list_of_cascade_models()
        **kwargs: Extra keyword arguments used for the initialization of the model

    Returns:
        Either a cascade model or a composite model. In any case, a model that can be given to the fit_model function.
    """
    cml = CascadeModelsLoader()
    sml = CompositeModelsLoader()
    try:
        return cml.load(model_name, **kwargs)
    except ImportError:
        try:
            return sml.load(model_name, **kwargs)
        except ImportError:
            raise ValueError('The model with the name "{}" could not be found.'.format(model_name))


def get_meta_info(model_name):
    """Get the meta information of a particular model

    Args:
        model_name (str): One of the models from get_list_of_composite_models() or get_list_of_cascade_models()

    Returns:
        Either a cascade model or a composite model. In any case, a model that can be given to the fit_model function.
    """
    cml = CascadeModelsLoader()
    sml = CompositeModelsLoader()
    try:
        return cml.get_meta_info(model_name)
    except ImportError:
        try:
            return sml.get_meta_info(model_name)
        except ImportError:
            raise ValueError('The model with the name "{}" could not be found.'.format(model_name))


class ComponentBuilder(object):

    def __init__(self):
        """The base class for component builders.

        Component builders, together with ComponentConfig allow you to define components using class attributes.

        The idea is that the ComponentConfig contains class attributes defining the component and that the
        ComponentBuilder is able to create a class of the right type from the information in the component config.
        """

    def create_class(self, template):
        """Create a class of the right type given the information in the template.

        Args:
            template (ComponentConfig): the information as a component config

        Returns:
            class: the class of the right type
        """


def bind_function(func):
    """This decorator is for methods in ComponentConfigs that we would like to bind to the constructed component.

    Example suppose you want to inherit or overwrite a function in the constructed model, then in your template/config
    you should define the function and add @bind_function to it as a decorator, like this:

    .. code-block:: python

        # the class we want to create
        class MyGoal(object):
            def test(self):
                print('test')

        # the template class from which we want to construct a new MyGoal, note the @bind_function
        class MyConfig(ComponentConfig):
            @bind_function
            def test(self):
                super(MyGoal, self).test()
                print('test2')

    The component builder takes care to actually bind the new method to the final object.

    What this will do essentially is that it will add the property bind to the function. This should act as a
    flag indicating that that function should be bound.

    Args:
        func (python function): the function to bind to the build object
    """
    func._bind = True
    return func


def method_binding_meta(template, *bases):
    """Adds all bound functions from the ComponentConfig to the class being constructed.

     This returns a metaclass similar to the with_metaclass of the six library.

     Args:
         template (ComponentConfig): the component config with the bound_methods attribute which we will all add
            to the attributes of the to creating class.
     """
    class ApplyMethodBinding(type):
        def __new__(mcs, name, bases, attributes):
            attributes.update(template.bound_methods)
            return super(ApplyMethodBinding, mcs).__new__(mcs, name, bases, attributes)

    return with_metaclass(ApplyMethodBinding, *bases)


class ComponentConfigMeta(type):

    def __new__(mcs, name, bases, attributes):
        """A pre-processor for the components.

        On the moment this meta class does two things, first it adds all functions with the '_bind' property
        to the bound_methods list for binding them later to the constructed class. Second, it sets the 'name' attribute
        of the component to the class name if there is no name attribute defined.
        """
        result = super(ComponentConfigMeta, mcs).__new__(mcs, name, bases, attributes)
        bound_methods = {value.__name__: value for value in attributes.values() if hasattr(value, '_bind')}
        for base in bases:
            if hasattr(base, 'bound_methods'):
                for key, value in base.bound_methods.items():
                    if key not in bound_methods:
                        bound_methods.update({key: value})
        result.bound_methods = bound_methods

        if 'name' not in attributes:
            result.name = name

        return result


class ComponentConfig(with_metaclass(ComponentConfigMeta, object)):
    """The component configuration.

    By overriding the class attributes you can define complex configurations. The actual class distilled from these
    configurations are loaded by the ComponentBuilder
    """
    name = ''
    description = ''

    @classmethod
    def meta_info(cls):
        return {'name': cls.name,
                'description': cls.description}


class ComponentsLoader(object):

    def __init__(self, sources):
        """The base class for loading and displaying components.

        Args:
            sources (:class:`list`): the list of sources to use for loading the components
        """
        self._sources = sources
        self._check_unique_names()

    def list_all(self):
        """List the names of all the available components."""
        components = []
        for source in self._sources:
            components.extend(source.list())
        return list(sorted(set(components)))

    def has_component(self, name):
        """Check if this loader has a component with the given name.

        Args:
            name (str): the name of the component

        Returns:
            boolean: true if this loader has the given component, false otherwise
        """
        try:
            self._get_preferred_source(name)
            return True
        except ImportError:
            return False

    def get_all_meta_info(self):
        """Get the meta information of all loadable components.

        Returns:
            dict: the keys are the names of the objects, as returned by list_all() and the values are the
                meta-information dicts.
        """
        return {k: self.get_meta_info(k) for k in self.list_all()}

    def get_meta_info(self, name):
        """Get the meta information of a component of the given name.

        Args:
            name (str): The name of the component we want to use

        Returns:
            dict: a dictionary with meta information for this component. Standard meta information is:
                - name (str): the name of the component
                - description (str): the description of the component
        """
        source = self._get_preferred_source(name)
        return source.get_meta_info(name)

    def get_class(self, name):
        """Get the class to the component of the given name.

        Args:
            name (str): The name of the component we want to use

        Returns:
            class or cb function to construct the given component
        """
        source = self._get_preferred_source(name)
        return source.get_class(name)

    def load(self, name, *args, **kwargs):
        """Load the component with the given name

        Args:
            name (str): The name of the component we want to use
            *args: passed to the component
            **kwargs: passed to the component

        Returns:
            the loaded module
        """
        c = self.get_class(name)
        return c(*args, **kwargs)

    def _get_preferred_source(self, name):
        """Try to get the preferred source for the component with the given name.

        The order of the sources matter, the first source takes precedence over the latter ones and so forth.
        """
        for source in self._sources:
            try:
                source.get_class(name)
                return source
            except ImportError:
                pass
        raise ImportError("No component found with the name {}".format(name))

    def _check_unique_names(self):
        """Check if all the elements in the sources are unique."""
        elements = []
        for source in self._sources:
            elements.extend(source.list())
        non_unique = list(item for item, count in collections.Counter(elements).items() if count > 1)
        if len(non_unique):
            raise NonUniqueComponent('Non-unique components detected: {}, please rename them.'.format(non_unique))


class ComponentsSource(object):

    def __init__(self):
        """Defines a source for components.

        This has functions for listing the available components as well as getting the class and meta information.
        """

    def list(self):
        """Get the names of all the available components from this source.

        Returns:
            list or str: list of the names of all the components loadable from this source.
        """
        return []

    def get_class(self, name):
        """Get the class for the component by the given name

        Args:
            name (str): The name of the component we want to use

        Returns:
            the construction function
        """
        raise ImportError

    def get_meta_info(self, name):
        """Get the meta information of a component of the given name.

        Args:
            name (str): The name of the component we want to use

        Returns:
            dict: a dictionary with meta information for this component. Standard meta information is:
                - name (str): the name of the component
                - description (str): the description of the component
        """
        return {}


class UserComponentsSourceSingle(ComponentsSource):

    def __init__(self, user_type, component_type):
        """Load the user components of the type *single*.

        This expects that the available python files contain a class with the same name as the file in which the class
        resides. For example, if we have a python file named "MySingleComponent.py" we should at least have the class
        named "MySingleComponent" in that python file. Additionally, the file can contain a dictionary named
        'meta_info' which contains meta information about that module.

        Args:
            user_type (str): either 'standard' or 'user'
            component_type (str): the type of component we wish to use. This should be named exactly to one of the
                directories available in mdt/data/components/
        """
        super(UserComponentsSourceSingle, self).__init__()
        self.path = _get_components_path(user_type, component_type)
        self._class_filenames = {}

    def list(self):
        items = []
        if os.path.isdir(self.path):
            for dir_name, sub_dirs, files in os.walk(self.path):
                for file in files:
                    if file.endswith('.py') and not file.startswith('__'):
                        items.append(file[0:-3])
                        self._class_filenames[file[0:-3]] = os.path.join(dir_name, file)
        return items

    def get_class(self, name):
        if name in self._class_filenames:
            module = imp.load_source(name, self._class_filenames[name])
            return getattr(module, name)
        raise ImportError

    def get_meta_info(self, name):
        path = os.path.join(self.path, name + '.py')
        if os.path.exists(path):
            module = imp.load_source(name, path)
            try:
                return getattr(module, 'meta_info')
            except AttributeError:
                try:
                    cls = self.get_class(name)
                    return cls.meta_info()
                except AttributeError:
                    return {}
        return {}


class AutoUserComponentsSourceSingle(UserComponentsSourceSingle):

    def __init__(self, user_type, component_type, component_builder):
        """

        This class extends the default single components source loader by also being able to use components defined
        using the ComponentConfig method. This means that the components are defined as subclasses of ComponentConfig
        and we need a ComponentBuilder to actually create the components.

        Args:
            user_type (str): either 'standard' or 'user'
            component_type (str): the type of component we wish to use. This should be named exactly to one of the
                directories available in mdt/data/components/
            component_builder (ComponentBuilder): the component creator that can create components using
                ComponentConfig classes
        """
        self.component_builder = component_builder
        super(AutoUserComponentsSourceSingle, self).__init__(user_type, component_type)

    def get_class(self, name):
        cls = super(AutoUserComponentsSourceSingle, self).get_class(name)
        if issubclass(cls, ComponentConfig):
            return self.component_builder.create_class(cls)
        return cls


class ComponentInfo(object):

    def get_name(self):
        """Get the name of this component

        Returns:
            str: the name of this component
        """
        return NotImplemented

    def get_meta_info(self):
        """Get the additional meta info of this component

        Returns:
            dict: the meta info
        """
        return NotImplemented

    def get_component_class(self):
        """Get the component class

        Returns:
            class: the class of the component
        """
        return NotImplemented


class UserComponentsSourceMulti(ComponentsSource):

    loaded_modules_cache = {}

    def __init__(self, user_type, component_type):
        """"Base class for components in which there are multiple components per file.

        Args:
            user_type (str): either 'user' or 'standard'. This defines from which dir to use the components
            component_type (str): from which dir in 'user' or 'standard' to use the components

        Attributes:
            loaded_modules_cache (dict): A cache for loaded components.
                If we do not do this, we fit_model into TypeError problems when a class is reloaded while there is
                already an instantiated object.

                This dict is indexed per directory names and contains for each loaded python file a tuple with the
                module and the loaded component.
        """
        super(UserComponentsSourceMulti, self).__init__()
        self._user_type = user_type
        self._component_type = component_type

        if self._user_type not in self.loaded_modules_cache:
            self.loaded_modules_cache[self._user_type] = {}

        if self._component_type not in self.loaded_modules_cache[self._user_type]:
            self.loaded_modules_cache[self._user_type][self._component_type] = {}

        self.path = _get_components_path(user_type, component_type)
        self._check_path()
        self._components = self._load_all_components()

    def list(self):
        return self._components.keys()

    def get_class(self, name):
        if name not in self._components:
            raise ImportError
        return self._components[name].get_component_class()

    def get_meta_info(self, name):
        return self._components[name].get_meta_info()

    def _load_all_components(self):
        self._update_modules_cache()

        all_components = []
        for module, components in self.loaded_modules_cache[self._user_type][self._component_type].values():
            all_components.extend(components)

        return {component.get_name(): component for component in all_components}

    def _update_modules_cache(self):
        """Fill the modules cache with the components.

        This loops through all the python files present in the dir name and tries to use them as modules if they
        are not yet loaded.
        """
        for dir_name, sub_dirs, files in os.walk(self.path):
            for file in files:
                if file.endswith('.py') and not file.startswith('__'):
                    path = os.path.join(dir_name, file)

                    if path not in self.loaded_modules_cache[self._user_type][self._component_type]:
                        module_name = self._user_type + '/' + \
                                      self._component_type + '/' + \
                                      dir_name[len(self.path) + 1:] + '/' + \
                                      os.path.splitext(os.path.basename(path))[0]

                        module = imp.load_source(module_name, path)

                        self.loaded_modules_cache[self._user_type][self._component_type][path] = \
                            (module, self._get_components_from_module(module))

    def _get_components_from_module(self, module):
        """Return a list of all the available components in the given module.

        Args:
            module (module): the module from which to use the components.

        Returns:
            list: list of ComponentInfo objects
        """
        loaded_items = []

        if hasattr(module, 'get_components_list'):
            loaded_items.extend(module.get_components_list())

        return loaded_items

    def _check_path(self):
        if not os.path.isdir(self.path):
            raise RuntimeError('The components folder "{0}" could not be found. '
                               'Please check the path to the components in your configuration file.'.format(self.path))


class AutoUserComponentsSourceMulti(UserComponentsSourceMulti):

    def __init__(self, user_type, component_type, component_class, component_builder):
        """Create a component source that can create components using multiple types of definitions.

        This will use either objects of the class defined by component_class or it will use objects
        of type ComponentConfig using the builder defined by component_creator or it will use the objects
        from the get_components_list function.

        Args:
            user_type (str): either 'user' or 'standard'. This defines from which dir to use the components
            component_type (str): from which dir in 'user' or 'standard' to use the components
            component_class (class): the class to auto use
            component_builder (ComponentBuilder): the component creator to use for components defined as a
                ComponentConfig.
        """
        self._component_class = component_class
        self.component_builder = component_builder
        super(AutoUserComponentsSourceMulti, self).__init__(user_type, component_type)

    def get_class(self, name):
        if name not in self._components:
            raise ImportError

        base = self._components[name].get_component_class()
        if inspect.isclass(base) and issubclass(base, ComponentConfig):
            return self.component_builder.create_class(base)

        return super(AutoUserComponentsSourceMulti, self).get_class(name)

    def _get_components_from_module(self, module):
        """Return a list of all the available components in the given module.

        Args:
            module (module): the module from which to use the components.

        Returns:
            list: list of components loaded from this module
        """
        class DynamicInfo(ComponentInfo):

            def __init__(self, component):
                self._component = component

            def get_name(self):
                return self._component.name

            def get_meta_info(self):
                return self._component.meta_info()

            def get_component_class(self):
                return self._component

        loaded_items = super(AutoUserComponentsSourceMulti, self)._get_components_from_module(module)

        items = inspect.getmembers(module, _get_class_predicate(module, self._component_class))
        loaded_items.extend(DynamicInfo(item[1]) for item in items)

        items = inspect.getmembers(module, _get_class_predicate(module, ComponentConfig))
        loaded_items.extend(DynamicInfo(item[1]) for item in items)

        return loaded_items


class ParametersSource(AutoUserComponentsSourceMulti):

    def __init__(self, user_type):
        """Source for the items in the 'parameters' dir in the components folder."""
        from mot.model_building.cl_functions.parameters import CLFunctionParameter
        from mdt.models.parameters import ParameterBuilder
        super(ParametersSource, self).__init__(user_type, 'parameters', CLFunctionParameter, ParameterBuilder())


class CompositeModelSource(AutoUserComponentsSourceMulti):

    def __init__(self, user_type):
        """Source for the items in the 'composite_models' dir in the components folder."""
        from mdt.models.composite import DMRICompositeModel, DMRICompositeModelBuilder
        super(CompositeModelSource, self).__init__(user_type, 'composite_models', DMRICompositeModel,
                                                DMRICompositeModelBuilder())


class CascadeSource(AutoUserComponentsSourceMulti):

    def __init__(self, user_type):
        """Source for the items in the 'cascade_models' dir in the components folder."""
        from mdt.models.cascade import DMRICascadeModelInterface, CascadeBuilder
        super(CascadeSource, self).__init__(user_type, 'cascade_models', DMRICascadeModelInterface, CascadeBuilder())


class MOTSourceSingle(ComponentsSource):

    def get_meta_info(self, name):
        return {}


class MOTLibraryFunctionSource(MOTSourceSingle):

    def get_class(self, name):
        return getattr(mot.model_building.cl_functions.library_functions, name)

    def list(self):
        module = mot.model_building.cl_functions.library_functions
        items = inspect.getmembers(module,  _get_class_predicate(module, LibraryFunction))
        return [x[0] for x in items if x[0] != 'LibraryFunction']


class MOTCompartmentModelsSource(MOTSourceSingle):

    def get_class(self, name):
        return getattr(mot.model_building.cl_functions.model_functions, name)

    def list(self):
        module = mot.model_building.cl_functions.model_functions
        items = inspect.getmembers(module, _get_class_predicate(module, ModelFunction))
        return [x[0] for x in items if x[0] != 'ModelFunction']


class BatchProfilesLoader(ComponentsLoader):

    def __init__(self):
        super(BatchProfilesLoader, self).__init__([UserComponentsSourceSingle('user', 'batch_profiles'),
                                                   UserComponentsSourceSingle('standard', 'batch_profiles')])


class ProcessingStrategiesLoader(ComponentsLoader):

    def __init__(self):
        super(ProcessingStrategiesLoader, self).__init__(
            [UserComponentsSourceSingle('user', 'processing_strategies'),
             UserComponentsSourceSingle('standard', 'processing_strategies')])


class NoiseSTDCalculatorsLoader(ComponentsLoader):

    def __init__(self):
        super(NoiseSTDCalculatorsLoader, self).__init__(
            [UserComponentsSourceSingle('user', 'noise_std_estimators'),
             UserComponentsSourceSingle('standard', 'noise_std_estimators')])


class CompartmentModelsLoader(ComponentsLoader):

    def __init__(self):
        from mdt.models.compartments import CompartmentBuilder
        super(CompartmentModelsLoader, self).__init__(
            [AutoUserComponentsSourceSingle('user', 'compartment_models', CompartmentBuilder()),
             AutoUserComponentsSourceSingle('standard', 'compartment_models', CompartmentBuilder()),
             MOTCompartmentModelsSource()])


class LibraryFunctionsLoader(ComponentsLoader):

    def __init__(self):
        super(LibraryFunctionsLoader, self).__init__([UserComponentsSourceSingle('user', 'library_functions'),
                                                      UserComponentsSourceSingle('standard', 'library_functions'),
                                                      MOTLibraryFunctionSource()])


class CompositeModelsLoader(ComponentsLoader):

    def __init__(self):
        super(CompositeModelsLoader, self).__init__([CompositeModelSource('user'),
                                                     CompositeModelSource('standard')])


class ParametersLoader(ComponentsLoader):

    def __init__(self):
        super(ParametersLoader, self).__init__([ParametersSource('user'),
                                                ParametersSource('standard')])


class CascadeModelsLoader(ComponentsLoader):

    def __init__(self):
        super(CascadeModelsLoader, self).__init__([CascadeSource('user'),
                                                   CascadeSource('standard')])


def get_component_class(component_type, component_name):
    """Return the class of the given component.

    Args:
        component_type (str): the type of component, for example 'batch_profiles' or 'parameters'
        component_name (str): the name of the component to use

    Returns:
        the class of the given component
    """
    if component_type == 'batch_profiles':
        return BatchProfilesLoader().get_class(component_name)
    if component_type == 'cascade_models':
        return CascadeModelsLoader().get_class(component_name)
    if component_type == 'compartment_models':
        return CompartmentModelsLoader().get_class(component_name)
    if component_type == 'library_functions':
        return LibraryFunctionsLoader().get_class(component_name)
    if component_type == 'noise_std_estimators':
        return NoiseSTDCalculatorsLoader().get_class(component_name)
    if component_type == 'parameters':
        return ParametersLoader().get_class(component_name)
    if component_type == 'processing_strategies':
        return ProcessingStrategiesLoader().get_class(component_name)
    if component_type == 'composite_models':
        return CompositeModelsLoader().get_class(component_name)
    raise ValueError('Could not find the given component type {}'.format(component_type))


def load_component(component_type, component_name, *args, **kwargs):
    """Load the class indicated by the given component type and name.

    Args:
        component_type (str): the type of component, for example 'batch_profiles' or 'parameters'
        component_name (str): the name of the component to use
        *args: passed to the component
        **kwargs: passed to the component

    Returns:
        the loaded component
    """
    component = get_component_class(component_type, component_name)
    return component(*args, **kwargs)


def _get_class_predicate(module, class_type):
    """A predicate to be used in the function inspect.getmembers

    This predicate checks if the module of the item we inspect matches the given module, checks the class type
    to be the given class type and checks if the checked item is not in the exclude list.

    Args:
        module (module): the module to check against the module of the item
        class_type (module): the module to check against the class_type of the item

    Returns:
        function: a function to be used as a predicate in inspect.getmembers
    """
    def defined_in_module(item):
        return item.__module__ == module.__name__

    def complete_predicate(item):
        return inspect.isclass(item) and defined_in_module(item) and issubclass(item, class_type)

    return complete_predicate


def _get_components_path(user_type, component_type):
    """
    Args:
        user_type (str): either 'standard' or 'user'
        component_type (str): one of the dir names in standard and user
    """
    from mdt.configuration import get_config_dir
    return os.path.join(get_config_dir(), 'components', user_type, component_type)



import time
from contextlib import contextmanager
from functools import wraps
from PyQt5.QtCore import QObject, pyqtSignal, QFileSystemWatcher, pyqtSlot, QTimer
from PyQt5.QtCore import QTimer
from PyQt5.QtWidgets import QApplication

from mdt.nifti import yield_nifti_info
from mdt.log_handlers import LogListenerInterface

__author__ = 'Robbert Harms'
__date__ = "2015-08-20"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


class QtManager(object):

    windows = []

    @staticmethod
    def get_qt_application_instance():
        q_app = QApplication.instance()
        if q_app is None:
            q_app = QApplication([])
            q_app.lastWindowClosed.connect(QtManager.empty_windows_list)
        return q_app

    @staticmethod
    def exec_():
        if QtManager.windows:
            QtManager.get_qt_application_instance().exec_()

    @staticmethod
    def add_window(window):
        QtManager.windows.append(window)

    @staticmethod
    def empty_windows_list():
        QtManager.windows = []


def center_window(window):
    """Center the given window on the screen.

    Args:
        q_app (QApplication): for desktop information
        window (QMainWindow): the window to center
    """
    q_app = QApplication.instance()

    frame_gm = window.frameGeometry()
    screen = q_app.desktop().screenNumber(q_app.desktop().cursor().pos())
    center_point = q_app.desktop().screenGeometry(screen).center()
    frame_gm.moveCenter(center_point)
    window.move(frame_gm.topLeft())


def function_message_decorator(header, footer):
    """This creates and returns a decorator that prints a header and footer before executing the function.

    Args:
        header (str): the header text, we will add extra decoration to it
        footer (str): the footer text, we will add extra decoration to it

    Returns:
        decorator function
    """
    def _called_decorator(dec_func):

        @wraps(dec_func)
        def _decorator(*args, **kwargs):
            print('')
            print(header)
            print('-'*20)

            response = dec_func(*args, **kwargs)

            print('-'*20)
            print(footer)

            return response
        return _decorator

    return _called_decorator


@contextmanager
def blocked_signals(*widgets):
    """Small context in which the signals of the given widget are blocked.

    Args:
        widgets (QWidget): one or more widgets
    """
    def apply_block(bool_val):
        for w in widgets:
            w.blockSignals(bool_val)

    apply_block(True)
    yield
    apply_block(False)


def print_welcome_message():
    """Prints a small welcome message for after the GUI has loaded.

    This prints to stdout. We expect the GUI to catch the stdout events and redirect them to the GUI.
    """
    from mdt import VERSION
    print('Welcome to MDT version {}.'.format(VERSION))
    print('')
    print('This area is reserved for log output.')
    print('-------------------------------------')


class ForwardingListener(LogListenerInterface):

    def __init__(self, queue):
        """Forwards all incoming messages to the given _logging_update_queue.

        Instances of this class can be used as a log listener to the MDT LogDispatchHandler and as a
        sys.stdout replacement.

        Args:
            queue (Queue): the _logging_update_queue to forward the messages to
        """
        self._queue = queue

    def emit(self, record, formatted_message):
        self._queue.put(formatted_message + "\n")

    def write(self, string):
        self._queue.put(string)


image_files_filters = ['Nifti (*.nii *.nii.gz)',
                       'IMG, HDR (*.img)',
                       'All files (*)']
protocol_files_filters = ['MDT protocol (*.prtcl)',
                          'Text files (*.txt)',
                          'All files (*)']


class UpdateDescriptor(object):

    def __init__(self, attribute_name):
        """Descriptor that will emit a state_updated_signal at each update.

        This accesses from the instance the attribute name prepended with an underscore (_).
        """
        self._attribute_name = attribute_name

    def __get__(self, instance, owner):
        return getattr(instance, '_' + self._attribute_name)

    def __set__(self, instance, value):
        setattr(instance, '_' + self._attribute_name, value)
        instance.state_updated_signal.emit(self._attribute_name)


class MessageReceiver(QObject):

    text_message_signal = pyqtSignal(str)
    finished = pyqtSignal()

    def __init__(self, queue, *args, **kwargs):
        """A QObject (to be run in a QThread) which sits waiting for data to come through a Queue.Queue().

        It blocks until data is available, and one it has got something from the _logging_update_queue, it sends
        it to the "MainThread" by emitting a Qt Signal.

        Attributes:
            is_running (boolean): set to False to stop the receiver.
        """
        super(MessageReceiver, self).__init__(*args, **kwargs)
        self.queue = queue
        self.is_running = True

    def run(self):
        while self.is_running:
            if not self.queue.empty():
                self.text_message_signal.emit(self.queue.get())
            time.sleep(0.001)
        self.finished.emit()


class MainTab(object):

    def tab_opened(self):
        """Called when this tab is selected by the user."""


class DirectoryImageWatcher(QObject):

    image_updates = pyqtSignal(tuple, tuple)

    def __init__(self, directory=None):
        """Watches a given directory for added, removed and/or updated nifti files.

        Args:
            directory (str): the initial directory to watch. You can set a new one with set_directory().

        Signals
            images_updates (list, list, dict): sent when images are updated in the watch directory. It contains
                the list of additions, removals and changes where changes is a dict mapping the old name to a new one.
        """
        super(DirectoryImageWatcher, self).__init__()
        self._watched_dir = None
        self._current_files = []
        self._watcher = QFileSystemWatcher()
        self._watcher.directoryChanged.connect(self._directory_changed)
        self._timer = QTimer()
        self._timer.timeout.connect(self._timer_event)
        self._timer.timeout.connect(self._timer.stop)

        if directory:
            self.set_directory(directory)

    def set_directory(self, directory):
        """Set the watched directory to the given directory.

        Args:
            directory (str): the new directory to watch for added, removed and/or changed nifti files.
        """
        if self._watched_dir:
            self._watcher.removePath(self._watched_dir)
        self._watched_dir = directory
        self._watcher.addPath(directory)
        self._current_files = list(el[1] for el in yield_nifti_info(directory))

    @pyqtSlot(str)
    def _directory_changed(self, directory):
        if directory == self._watched_dir:
            self._timer.start(100)

    @pyqtSlot()
    def _timer_event(self):
        new_file_list = list(el[1] for el in yield_nifti_info(self._watched_dir))

        removals = set(self._current_files).difference(new_file_list)
        additions = set(new_file_list).difference(self._current_files)

        self._current_files = new_file_list
        self.image_updates.emit(tuple(additions), tuple(removals))


class TimedUpdate(QTimer):

    def __init__(self, update_cb):
        """Creates a timer that can delay running a given callback function.

        Every time the user adds a delayed callback the timer gets reset to the new value and we will
        wait that new value until calling the callback with the last data given.

        Args:
            update_cb (function): the function we would like to run after a timer has run out
        """
        super(TimedUpdate, self).__init__()
        self._cb_values = []
        self._update_cb = update_cb
        self.timeout.connect(self._call_update_cb)
        self.timeout.connect(self.stop)

    def add_delayed_callback(self, delay, *cb_values):
        """Pushes a new delay to calling the callback function.

        Args:
            delay (int): the time in ms to wait
            cb_values (*list): the list of values to use as arguments to the callback function. Leave empty to disable.
        """
        self._cb_values = cb_values
        self.start(delay)

    def _call_update_cb(self):
        if self._cb_values:
            self._update_cb(*self._cb_values)
        else:
            self._update_cb()



from mdt.models.composite import DMRICompositeModelConfig

__author__ = 'Robbert Harms'
__date__ = "2015-06-22"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


class BallStick_r1(DMRICompositeModelConfig):

    ex_vivo_suitable = False
    description = 'The default Ball & Stick model'
    model_expression = '''
        S0 * ( (Weight(w_ball) * Ball) +
               (Weight(w_stick) * Stick) )
    '''
    fixes = {'Ball.d': 3.0e-9,
             'Stick.d': 1.7e-9}
    post_optimization_modifiers = [('FS', lambda results: 1 - results['w_ball.w'])]


class BallStick_r1_ExVivo(BallStick_r1):

    name = 'BallStick_r1-ExVivo'
    in_vivo_suitable = False
    ex_vivo_suitable = True
    description = 'The Ball & Stick model with ex vivo defaults',
    fixes = {'Ball.d': 2.0e-9,
             'Stick.d': 0.6e-9}


class BallStick_r2(DMRICompositeModelConfig):

    ex_vivo_suitable = False
    description = 'The Ball & 2x Stick model'
    model_expression = '''
        S0 * ( (Weight(w_ball) * Ball) +
               (Weight(w_stick0) * Stick(Stick0)) +
               (Weight(w_stick1) * Stick(Stick1)) )
    '''
    fixes = {'Ball.d': 3.0e-9,
             'Stick0.d': 1.7e-9,
             'Stick1.d': 1.7e-9}
    post_optimization_modifiers = [('FS', lambda results: 1 - results['w_ball.w'])]


class BallStick_r2_ExVivo(BallStick_r2):

    name = 'BallStick_r2-ExVivo'
    in_vivo_suitable = False
    ex_vivo_suitable = True
    description = 'The Ball & 2x Stick model with ex vivo defaults'
    fixes = {'Ball.d': 2.0e-9,
             'Stick0.d': 0.6e-9,
             'Stick1.d': 0.6e-9}


class BallStick_r3(DMRICompositeModelConfig):

    ex_vivo_suitable = False
    description = 'The Ball & 3x Stick model'
    model_expression = '''
            S0 * ( (Weight(w_ball) * Ball) +
                   (Weight(w_stick0) * Stick(Stick0)) +
                   (Weight(w_stick1) * Stick(Stick1)) +
                   (Weight(w_stick2) * Stick(Stick2)) )
        '''
    fixes = {'Ball.d': 3.0e-9,
             'Stick0.d': 1.7e-9,
             'Stick1.d': 1.7e-9,
             'Stick2.d': 1.7e-9}
    inits = {'w_stick2.w': 0}

    post_optimization_modifiers = [('FS', lambda results: 1 - results['w_ball.w'])]


class BallStick_r3_ExVivo(BallStick_r3):

    name = 'BallStick_r3-ExVivo'
    ex_vivo_suitable = True
    in_vivo_suitable = False
    description = 'The Ball & 3x Stick model with ex vivo defaults'
    fixes = {'Ball.d': 2.0e-9,
             'Stick0.d': 0.6e-9,
             'Stick1.d': 0.6e-9,
             'Stick2.d': 0.6e-9}



import os

from PyQt5.QtCore import pyqtSlot, QObject, pyqtSignal
from PyQt5.QtWidgets import QFileDialog, QMessageBox, QDialog, QDialogButtonBox
from mdt.gui.model_fit.design.ui_optimization_options_dialog import Ui_OptimizationOptionsDialog

import mdt
from mdt.gui.model_fit.design.ui_fit_model_tab import Ui_FitModelTabContent
from mdt.gui.utils import function_message_decorator, image_files_filters, protocol_files_filters, MainTab
from mdt.utils import split_image_path
from mot.factory import get_optimizer_by_name

__author__ = 'Robbert Harms'
__date__ = "2016-06-27"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


class FitModelTab(MainTab, Ui_FitModelTabContent):

    def __init__(self, shared_state, computations_thread):
        self._shared_state = shared_state
        self._computations_thread = computations_thread
        self._run_model_worker = RunModelWorker()
        self._tab_content = None
        self._optim_options = OptimOptions()

    def setupUi(self, tab_content):
        super(FitModelTab, self).setupUi(tab_content)
        self._tab_content = tab_content

        self.selectDWI.clicked.connect(lambda: self._select_dwi())
        self.selectMask.clicked.connect(lambda: self._select_mask())
        self.selectProtocol.clicked.connect(lambda: self._select_protocol())
        self.selectOutputFolder.clicked.connect(lambda: self._select_output())

        self.selectedDWI.textChanged.connect(self._check_enable_action_buttons)
        self.selectedMask.textChanged.connect(self._check_enable_action_buttons)
        self.selectedProtocol.textChanged.connect(self._check_enable_action_buttons)

        self.runButton.clicked.connect(lambda: self.run_model())
        self.optimizationOptionsButton.clicked.connect(self._run_optimization_options_dialog)

        self.modelSelection.addItems(list(sorted(mdt.get_models_list())))
        self.modelSelection.setCurrentText('BallStick_r1 (Cascade)')

        self._check_enable_action_buttons()

    def _select_dwi(self):
        initial_dir = self._shared_state.base_dir
        if self.selectedDWI.text() != '':
            initial_dir = self.selectedDWI.text()

        open_file, used_filter = QFileDialog().getOpenFileName(
            caption='Select the 4d diffusion weighted image', directory=initial_dir,
            filter=';;'.join(image_files_filters))

        if os.path.isfile(open_file):
            self.selectedDWI.setText(open_file)
            self._shared_state.base_dir = os.path.dirname(open_file)
            self.update_output_folder_text()

    def _select_mask(self):
        initial_dir = self._shared_state.base_dir
        if self.selectedMask.text() != '':
            initial_dir = self.selectedMask.text()

        open_file, used_filter = QFileDialog().getOpenFileName(
            caption='Select the brain mask', directory=initial_dir,
            filter=';;'.join(image_files_filters))

        if os.path.isfile(open_file):
            self.selectedMask.setText(open_file)
            self._shared_state.base_dir = os.path.dirname(open_file)
            self.update_output_folder_text()

    def _select_output(self):
        initial_dir = self._shared_state.base_dir
        if self.selectedOutputFolder.text() != '':
            initial_dir = self.selectedOutputFolder.text()

        output_file_name = QFileDialog().getExistingDirectory(
            caption='Select the output folder', directory=initial_dir)

        if output_file_name:
            self.selectedOutputFolder.setText(output_file_name)

    def _select_protocol(self):
        initial_dir = self._shared_state.base_dir
        if self.selectedProtocol.text() != '':
            initial_dir = self.selectedProtocol.text()

        open_file, used_filter = QFileDialog().getOpenFileName(
            caption='Select the protocol', directory=initial_dir,
            filter=';;'.join(protocol_files_filters))

        if os.path.isfile(open_file):
            self.selectedProtocol.setText(open_file)
            self._shared_state.base_dir = os.path.dirname(open_file)

    def _check_enable_action_buttons(self):
        self.runButton.setEnabled(
            os.path.isfile(self.selectedDWI.text()) and
            os.path.isfile(self.selectedMask.text()) and
            os.path.isfile(self.selectedProtocol.text()) and
            self.selectedOutputFolder.text() != '')

    def update_output_folder_text(self):
        if os.path.isfile(self.selectedDWI.text()) and os.path.isfile(self.selectedMask.text()):
            folder_base = os.path.join(os.path.dirname(self.selectedDWI.text()), 'output',
                                       split_image_path(self.selectedMask.text())[1])
            self.selectedOutputFolder.setText(folder_base)

    def _run_optimization_options_dialog(self):
        dialog = OptimizationOptionsDialog(self._shared_state, self._tab_content, self._optim_options)
        return_value = dialog.exec_()

        if return_value:
            dialog.write_config()

    @pyqtSlot()
    def run_model(self):
        model = mdt.get_model(self.modelSelection.currentText())
        protocol = mdt.load_protocol(self.selectedProtocol.text())

        if not model.is_protocol_sufficient(protocol):
            msg = ProtocolWarningBox(model.get_protocol_problems(protocol))
            msg.exec_()
            return

        self._run_model_worker.set_args(
            model,
            mdt.load_problem_data(self.selectedDWI.text(),
                                  self.selectedProtocol.text(),
                                  self.selectedMask.text(),
                                  noise_std=self._optim_options.noise_std),
            self.selectedOutputFolder.text(),
            recalculate=True,
            double_precision=self._optim_options.double_precision,
            only_recalculate_last=not self._optim_options.recalculate_all,
            optimizer=self._optim_options.get_optimizer(),
            save_user_script_info=False)

        self._computations_thread.start()
        self._run_model_worker.moveToThread(self._computations_thread)

        self._run_model_worker.starting.connect(self._computations_thread.starting)
        self._run_model_worker.finished.connect(self._computations_thread.finished)

        self._run_model_worker.starting.connect(lambda: self.runButton.setEnabled(False))
        self._run_model_worker.finished.connect(lambda: self.runButton.setEnabled(True))

        self._run_model_worker.finished.connect(
            lambda: self._shared_state.set_output_folder(self._get_full_model_output_path()))

        self._run_model_worker.starting.emit()

    def _get_full_model_output_path(self):
        parts = [self.selectedOutputFolder.text()]
        parts.append(self.modelSelection.currentText().split(' ')[0])
        return os.path.join(*parts)


class ProtocolWarningBox(QMessageBox):

    def __init__(self, problems, *args):
        super(ProtocolWarningBox, self).__init__(*args)
        self.setIcon(QMessageBox.Warning)
        self.setWindowTitle("Insufficient protocol")
        self.setText("The provided protocol is insufficient for this model.")
        self.setInformativeText("The reported problems are: \n{}".format('\n'.join(' - ' + str(p) for p in problems)))
        self._in_resize = False

    def resizeEvent(self, event):
        if not self._in_resize:
            self._in_resize = True
            self.setFixedWidth(self.children()[-1].size().width() + 200)
            self._in_resize = False


class OptimizationOptionsDialog(Ui_OptimizationOptionsDialog, QDialog):

    def __init__(self, shared_state, parent, config):
        super(OptimizationOptionsDialog, self).__init__(parent)
        self._shared_state = shared_state
        self._config = config
        self.setupUi(self)

        self.noiseStdFileSelect.clicked.connect(lambda: self._select_std_file())
        self.noiseStd.textChanged.connect(self._check_enable_ok_button)
        self.patience.textChanged.connect(self._check_enable_ok_button)

        self.optimizationRoutine.addItems(sorted(OptimOptions.optim_routines.keys()))
        self.optimizationRoutine.currentIndexChanged.connect(self._update_default_patience)
        self.defaultOptimizerGroup.buttonClicked.connect(self._update_optimization_routine_selection)

        self._load_config()

    def write_config(self):
        """Write to the config the user selected options"""
        noise_std_value = self.noiseStd.text()
        if noise_std_value == '':
            self._config.noise_std = None
        else:
            self._config.noise_std = noise_std_value
            try:
                self._config.noise_std = float(noise_std_value)
            except ValueError:
                pass

        self._config.double_precision = self.doublePrecision.isChecked()
        self._config.recalculate_all = self.recalculateAll_True.isChecked()
        self._config.use_model_default_optimizer = self.defaultOptimizer_True.isChecked()
        self._config.optimizer = OptimOptions.optim_routines[self.optimizationRoutine.currentText()]
        self._config.patience = int(self.patience.text())

    def _load_config(self):
        """Load the settings from the config into the GUI"""
        if self._config.noise_std is not None:
            self.noiseStd.setText(str(self._config.noise_std))
        self.doublePrecision.setChecked(self._config.double_precision)
        self.recalculateAll_True.setChecked(self._config.recalculate_all)
        self.defaultOptimizer_False.setChecked(not self._config.use_model_default_optimizer)
        self._update_optimization_routine_selection()

        self.optimizationRoutine.setCurrentText({v: k for k, v in
                                                 OptimOptions.optim_routines.items()}[self._config.optimizer])
        self.patience.setText(str(self._config.patience))

    def _select_std_file(self):
        open_file, used_filter = QFileDialog().getOpenFileName(
            caption='Select a noise std volume', directory=self._shared_state.base_dir,
            filter=';;'.join(image_files_filters))

        if open_file:
            self._shared_state.base_dir = os.path.dirname(open_file)
            self.noiseStd.setText(open_file)

    def _check_enable_ok_button(self):
        noise_std_value = self.noiseStd.text()
        noise_std_value_is_float = False
        try:
            float(noise_std_value)
            noise_std_value_is_float = True
        except ValueError:
            pass

        enabled = noise_std_value == '' or noise_std_value_is_float or os.path.isfile(noise_std_value)

        if self.defaultOptimizer_False.isChecked():
            try:
                int(self.patience.text())
            except ValueError:
                enabled = False

        self.buttonBox.button(QDialogButtonBox.Ok).setEnabled(enabled)

    def _update_optimization_routine_selection(self):
        self.optimizationRoutine.setDisabled(self.defaultOptimizer_True.isChecked())
        self.patience.setDisabled(self.defaultOptimizer_True.isChecked())

    def _update_default_patience(self):
        optimizer = get_optimizer_by_name(OptimOptions.optim_routines[self.optimizationRoutine.currentText()])
        self.patience.setText(str(optimizer.default_patience))


class OptimOptions(object):

    optim_routines = {'Powell\'s method': 'Powell',
                      'Nelder-Mead Simplex': 'NMSimplex',
                      'Levenberg Marquardt': 'LevenbergMarquardt'}

    def __init__(self):
        """Storage class for communication between the options dialog and the main frame"""
        self.use_model_default_optimizer = True
        self.double_precision = False

        self.optimizer = mdt.configuration.get_general_optimizer_name()
        self.patience = mdt.configuration.get_general_optimizer_settings()['patience']

        if self.patience is None:
            self.patience = get_optimizer_by_name(self.optimizer).default_patience

        self.recalculate_all = False
        self.noise_std = None

    def get_optimizer(self):
        if self.use_model_default_optimizer:
            return None
        optimizer = get_optimizer_by_name(self.optimizer)
        return optimizer(patience=self.patience)


class RunModelWorker(QObject):

    starting = pyqtSignal()
    finished = pyqtSignal()

    def __init__(self):
        super(RunModelWorker, self).__init__()
        self.starting.connect(self.run)
        self._args = []
        self._kwargs = {}

    def set_args(self, *args, **kwargs):
        self._args = args
        self._kwargs = kwargs

    @function_message_decorator('Starting model fitting, please wait.',
                                'Finished model fitting. You can view the results using the "View results" tab.')
    @pyqtSlot()
    def run(self):
        mdt.fit_model(*self._args, **self._kwargs)
        self.finished.emit()




from mdt.models.cascade import CascadeConfig

__author__ = 'Robbert Harms'
__date__ = "2015-06-22"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


class CHARMED1(CascadeConfig):

    name = 'CHARMED_r1 (Cascade)'
    description = 'Initializes the directions to Ball & Stick.'
    models = ('BallStick_r1 (Cascade)',
              'CHARMED_r1')
    inits = {'CHARMED_r1': [('CHARMEDRestricted0.theta', 'Stick.theta'),
                            ('CHARMEDRestricted0.phi', 'Stick.phi'),
                            ('Tensor.theta', 'Stick.theta'),
                            ('Tensor.phi', 'Stick.phi'),
                            ('w_res0.w', 'w_stick.w')]}


class CHARMEDR1S0(CascadeConfig):

    name = 'CHARMED_r1 (Cascade|S0)'
    description = 'Cascade for CHARMED r1 initialized with only an S0 fit.'
    models = ('S0',
              'CHARMED_r1')


class CHARMEDR1Fixed(CascadeConfig):

    name = 'CHARMED_r1 (Cascade|fixed)'
    description = 'Fixes the directions to Ball & Stick.'
    models = ('BallStick_r1 (Cascade)',
              'CHARMED_r1')
    inits = {'CHARMED_r1': [('Tensor.theta', 'Stick.theta'),
                            ('Tensor.phi', 'Stick.phi'),
                            ('w_res0.w', 'w_stick.w')]}
    fixes = {'CHARMED_r1': [('CHARMEDRestricted0.theta', 'Stick.theta'),
                            ('CHARMEDRestricted0.phi', 'Stick.phi')]}


class CHARMEDR2(CascadeConfig):

    name = 'CHARMED_r2 (Cascade)'
    description = 'Initializes the directions to 2x Ball & Stick.'
    models = ('BallStick_r2 (Cascade)',
              'CHARMED_r2')
    inits = {'CHARMED_r2': [('Tensor.theta', 'Stick0.theta'),
                            ('Tensor.phi', 'Stick0.phi'),
                            ('CHARMEDRestricted0.theta', 'Stick0.theta'),
                            ('CHARMEDRestricted0.phi', 'Stick0.phi'),
                            ('CHARMEDRestricted1.theta', 'Stick1.theta'),
                            ('CHARMEDRestricted1.phi', 'Stick1.phi'),
                            ('w_res0.w', 'w_stick0.w'),
                            ('w_res1.w', 'w_stick1.w')]}


class CHARMEDR2S0(CascadeConfig):

    name = 'CHARMED_r2 (Cascade|S0)'
    description = 'Initializes with only an S0 fit.'
    models = ('S0',
              'CHARMED_r2')


class CHARMEDR2Fixed(CascadeConfig):

    name = 'CHARMED_r2 (Cascade|fixed)'
    description = 'Fixes the directions to 2x Ball & Stick.'
    models = ('BallStick_r2 (Cascade)',
              'CHARMED_r2')
    inits = {'CHARMED_r2': [('Tensor.theta', 'Stick0.theta'),
                            ('Tensor.phi', 'Stick0.phi'),
                            ('w_res0.w', 'w_stick0.w'),
                            ('w_res1.w', 'w_stick1.w')]}
    fixes = {'CHARMED_r2': [('CHARMEDRestricted0.theta', 'Stick0.theta'),
                            ('CHARMEDRestricted0.phi', 'Stick0.phi'),
                            ('CHARMEDRestricted1.theta', 'Stick1.theta'),
                            ('CHARMEDRestricted1.phi', 'Stick1.phi'),
                            ]}


class CHARMED_r3(CascadeConfig):

    name = 'CHARMED_r3 (Cascade)'
    description = 'Initializes the directions to 3x Ball & Stick.'
    models = ('BallStick_r3 (Cascade)',
              'CHARMED_r3')
    inits = {'CHARMED_r3': [('Tensor.theta', 'Stick0.theta'),
                            ('Tensor.phi', 'Stick0.phi'),
                            ('w_res0.w', 'w_stick0.w'),
                            ('w_res1.w', 'w_stick1.w'),
                            ('w_res2.w', 'w_stick2.w'),
                            ('CHARMEDRestricted0.theta', 'Stick0.theta'),
                            ('CHARMEDRestricted0.phi', 'Stick0.phi'),
                            ('CHARMEDRestricted1.theta', 'Stick1.theta'),
                            ('CHARMEDRestricted1.phi', 'Stick1.phi'),
                            ('CHARMEDRestricted2.theta', 'Stick2.theta'),
                            ('CHARMEDRestricted2.phi', 'Stick2.phi'),
                            ]}


class CHARMED_r3_S0(CascadeConfig):

    name = 'CHARMED_r3 (Cascade|S0)'
    description = 'Initializes with only an S0 fit.'
    models = ('S0',
              'CHARMED_r3')


class CHARMED_r3_Fixed(CascadeConfig):

    name = 'CHARMED_r3 (Cascade|fixed)'
    description = 'Fixes the directions to 3x Ball & Stick.'
    models = ('BallStick_r3 (Cascade)',
              'CHARMED_r3')
    inits = {'CHARMED_r3': [('Tensor.theta', 'Stick0.theta'),
                            ('Tensor.phi', 'Stick0.phi'),
                            ('w_res0.w', 'w_stick0.w'),
                            ('w_res1.w', 'w_stick1.w'),
                            ('w_res2.w', 'w_stick2.w')]}
    fixes = {'CHARMED_r3': [('CHARMEDRestricted0.theta', 'Stick0.theta'),
                            ('CHARMEDRestricted0.phi', 'Stick0.phi'),
                            ('CHARMEDRestricted1.theta', 'Stick1.theta'),
                            ('CHARMEDRestricted1.phi', 'Stick1.phi'),
                            ('CHARMEDRestricted2.theta', 'Stick2.theta'),
                            ('CHARMEDRestricted2.phi', 'Stick2.phi')]}



#!/usr/bin/env python
# PYTHON_ARGCOMPLETE_OK
"""Generate the bval and bvec file from a protocol file."""
import argparse
import os
import mdt
from argcomplete.completers import FilesCompleter
import textwrap
import mdt.protocols
from mdt.shell_utils import BasicShellApplication
from mdt.protocols import write_bvec_bval

__author__ = 'Robbert Harms'
__date__ = "2015-08-18"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


class GenerateBvecBval(BasicShellApplication):

    def __init__(self):
        mdt.init_user_settings(pass_if_exists=True)

    def _get_arg_parser(self):
        description = textwrap.dedent(__doc__)
        description += self._get_citation_message()

        epilog = textwrap.dedent("""
        Examples of use:
            mdt-generate-bvec-bval my_protocol.prtcl
            mdt-generate-bvec-bval my_protocol.prtcl bvec_name.bvec bval_name.bval
        """)

        parser = argparse.ArgumentParser(description=description, epilog=epilog,
                                         formatter_class=argparse.RawTextHelpFormatter)

        parser.add_argument('protocol', help='the protocol file').completer = FilesCompleter()
        parser.add_argument('bvec', help="the output bvec file", nargs='?', default=None).completer = FilesCompleter()
        parser.add_argument('bval', help="the output bvec file", nargs='?', default=None).completer = FilesCompleter()

        return parser

    def run(self, args):
        protocol_base = os.path.join(os.path.dirname(os.path.realpath(args.protocol)),
                                     os.path.splitext(os.path.basename(args.protocol))[0])

        if args.bvec:
            bvec = os.path.realpath(args.bvec)
        else:
            bvec = protocol_base + '.bvec'

        if args.bval:
            bval = os.path.realpath(args.bval)
        else:
            bval = protocol_base + '.bval'

        write_bvec_bval(mdt.load_protocol(os.path.realpath(args.protocol)), bvec, bval)


if __name__ == '__main__':
    GenerateBvecBval().start()



from copy import deepcopy
import six
import mdt
from mdt.components_loader import ComponentConfig, ComponentBuilder, bind_function, method_binding_meta, get_meta_info
from mdt.model_protocol_problem import NamedProtocolProblem
from mdt.models.base import DMRIOptimizable

__author__ = 'Robbert Harms'
__date__ = "2015-04-24"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


class DMRICascadeModelInterface(DMRIOptimizable):

    def __init__(self, *args, **kwargs):
        """The interface to cascade models.

        A cascade model is a model consisting of multi-compartment models or other cascade models. The idea is that
        it contains a number of models that are to be optimized one after another with the output results of
        the previous fit used to initialize the next model.
        """
        super(DMRICascadeModelInterface, self).__init__(*args, **kwargs)
        self.double_precision = False

    @property
    def name(self):
        """Get the name of this cascade model.

        Returns:
            str: The name of this cascade model
        """
        return ''

    def has_next(self):
        """Check if this cascade model has a next model.

        Returns:
            boolean: True if there is a next model, false otherwise.
        """

    def get_next(self, output_previous_models):
        """Get the next model in the cascade. This is the only function called by the cascade model optimizer

        This class is supposed to remember which model is next.

        Args:
            output_previous_models (dict): The output of all the previous models. The first level of the
                dict is for the models and is indexed by the model name. The second layer contains all the maps.

        Returns:
            SampleModelInterface: The sample model used for the next fit_model.
        """

    def reset(self):
        """Reset the iteration over the cascade.

        The implementing class should now reset the iteration such that get_next gets the first model again.
        """

    def get_model(self, name):
        """Get one of the models in the cascade by name.

        Args:
            name (str): the name of the model we want to return

        Returns:
            the model we want to have or None if no model found
        """

    def get_model_names(self):
        """Get the names of the models in this cascade in order of execution.

        Returns:
            list of str: the names of the models in this list
        """

    def set_problem_data(self, problem_data):
        """Set the problem data in every model in the cascade."""


class SimpleCascadeModel(DMRICascadeModelInterface):

    def __init__(self, name, model_list):
        """Create a new cascade model from a given list of models.

        This class adds some standard bookkeeping to make implementing cascade models easier.

        Args:
            name (str): the name of this cascade model
            model_list (list of models): the list of models this cascade consists of
        """
        super(DMRICascadeModelInterface, self).__init__()
        self._name = name
        self._model_list = model_list
        self._iteration_position = 0
        self.problems_to_analyze = None

    @property
    def name(self):
        return self._name

    def has_next(self):
        return self._iteration_position != len(self._model_list)

    def get_next(self, output_previous_models):
        next_model = self._model_list[self._iteration_position]

        output_previous = {}

        if self._iteration_position > 0:
            previous_model = self._model_list[self._iteration_position - 1]
            output_previous = output_previous_models[previous_model.name]

        self._prepare_model(next_model, output_previous, output_previous_models)
        self._iteration_position += 1
        return self._set_model_options(next_model)

    def reset(self):
        self._iteration_position = 0

    def is_protocol_sufficient(self, protocol=None):
        return all(model.is_protocol_sufficient(protocol) for model in self._model_list)

    def get_protocol_problems(self, protocol=None):
        problems = []
        for model in self._model_list:
            problems.extend(map(lambda p: NamedProtocolProblem(p, model.name), model.get_protocol_problems(protocol)))
        return problems

    def get_required_protocol_names(self):
        protocol_names = []
        for model in self._model_list:
            protocol_names.extend(model.get_required_protocol_names())
        return list(set(protocol_names))

    def get_model(self, name):
        for model in self._model_list:
            if model.name == name:
                return self._set_model_options(model)
        return None

    def get_model_names(self):
        return [model.name for model in self._model_list]

    def set_problem_data(self, problem_data):
        for model in self._model_list:
            model.set_problem_data(problem_data)

    def _set_model_options(self, model):
        """The final hook before we return a model from this class.

        This can set all kind of additional extra's to the model before we return it using any of the functions
        in this class

        Args:
            model: the model to which we want to set the final functions

        Returns:
            model: the same model with all extra's set.
        """
        if self.problems_to_analyze:
            model.problems_to_analyze = self.problems_to_analyze
        model.double_precision = self.double_precision
        return model

    def _prepare_model(self, model, output_previous, output_all_previous):
        """Prepare the next model with the output of the previous model.

        By default this model initializes all parameter maps to the output of the previous model.

        Args:
            model: The model to prepare
            output_previous (dict): the output of the (direct) previous model.
            output_all_previous (dict): The output of all the previous models. Indexed first by model name, second
                by full parameter name.

        Returns:
            None, preparing should happen in-place.
        """
        if not isinstance(model, DMRICascadeModelInterface) and output_previous:
            for key, value in output_previous.items():
                if model.has_parameter(key):
                    model.init(key, value)


class CascadeConfig(ComponentConfig):
    """The cascade config to inherit from.

    These configs are loaded on the fly by the CascadeBuilder.

    Attributes:
        name (str): the name of this cascade
        description (str): the description
        models (tuple): the list of models we wish to optimize (in that order) Example:

            .. code-block:: python

                models = ('BallStick (Cascade)', 'Charmed_r1')

        inits (dict): per model the initializations from the previous model. Example:

            .. code-block:: python

                inits = {'Charmed_r1': [
                            ('Tensor.theta', 'Stick.theta'),
                            ('Tensor.phi', 'Stick.phi'),
                            ('w_res0.w', lambda output_previous, output_all_previous: output_previous['w_stick.w'])
                            ]
                        }

            In this example the Charmed_r1 model in the cascade initializes its Tensor compartment with a previous
            Ball&Stick model and initializes the restricted compartment volume fraction with the Stick fraction.
            You can either provide a string matching the parameter name of the exact previous model, or provide
            callback function that accepts both a dict containing the previous model estimates
            and a dict containing all previous model estimates by model name and returns a single initialization map
            or value.

        fixes (dict): per model the fixations from the previous model. Example:

            .. code-block:: python

                fixes = {'Charmed_r1': [('CharmedRestricted0.theta', 'Stick.theta'),
                                        ('CharmedRestricted0.phi', 'Stick.phi')]}

            The syntax is similar to that of the inits attribute.

        lower_bounds (dict): per model the lower bounds to set using the results from the previous model
            Example:

            .. code-block:: python

                lower_bounds = {'Charmed_r1': [
                    ('S0.s0', lambda output_previous, output_all_previous: 2 * np.min(output_previous['S0.s0']))
                ]}

            The syntax is similar to that of the inits attribute.

        upper_bounds (dict): per model the upper bounds to set using the results from the previous model
            Example:

            .. code-block:: python

                upper_bounds = {'Charmed_r1': [
                    ('S0.s0', lambda output_previous, output_all_previous: 2 * np.max(output_previous['S0.s0']))
                ]}

            The syntax is similar to that of the inits attribute.
    """
    name = ''
    description = ''
    models = ()
    inits = {}
    fixes = {}
    lower_bounds = {}
    upper_bounds = {}

    @bind_function
    def _prepare_model_cb(self, model, output_previous, output_all_previous):
        """Finalize the preparation of the model in this callback.

        This is called at the end of the regular _prepare_model function defined in the SimpleCascadeModel and
        as implemented by the AutoCreatedCascadeModel.

        Use this if you want to control more of the initialization of the next model than only the inits and fixes.
        """


    @classmethod
    def meta_info(cls):
        meta_info = deepcopy(ComponentConfig.meta_info())
        meta_info.update({'name': cls.name,
                          'description': cls.description,
                          'in_vivo_suitable': get_meta_info(cls.models[len(cls.models) - 1])['in_vivo_suitable'],
                          'ex_vivo_suitable': get_meta_info(cls.models[len(cls.models) - 1])['ex_vivo_suitable']
                          })
        return meta_info


class CascadeBuilder(ComponentBuilder):

    def create_class(self, template):
        """Creates classes with as base class SimpleCascadeModel

        Args:
            template (CascadeConfig): the cascade config template to use for creating the class with the right init
                settings.
        """
        class AutoCreatedCascadeModel(method_binding_meta(template, SimpleCascadeModel)):

            def __init__(self, *args):
                new_args = [deepcopy(template.name),
                            list(map(mdt.get_model, template.models))]
                for ind, arg in args:
                    new_args[ind] = arg
                super(AutoCreatedCascadeModel, self).__init__(*new_args)

            def _prepare_model(self, model, output_previous, output_all_previous):
                super(AutoCreatedCascadeModel, self)._prepare_model(model, output_previous, output_all_previous)

                def parse_value(v):
                    if isinstance(v, six.string_types):
                        return output_previous[v]
                    elif hasattr(v, '__call__'):
                        return v(output_previous, output_all_previous)
                    return v

                for item in template.inits.get(model.name, {}):
                    model.init(item[0], parse_value(item[1]))

                for item in template.fixes.get(model.name, {}):
                    model.fix(item[0], parse_value(item[1]))

                for item in template.lower_bounds.get(model.name, {}):
                    model.set_lower_bound(item[0], parse_value(item[1]))

                for item in template.upper_bounds.get(model.name, {}):
                    model.set_upper_bound(item[0], parse_value(item[1]))

                self._prepare_model_cb(model, output_previous, output_all_previous)

        return AutoCreatedCascadeModel



from mdt.models.composite import DMRICompositeModelConfig

__author__ = 'Robbert Harms'
__date__ = "2015-06-22"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


class Tensor(DMRICompositeModelConfig):

    ex_vivo_suitable = False
    description = 'The standard Tensor model with in vivo defaults.'
    model_expression = '''
        S0 * Tensor
    '''
    inits = {'Tensor.d': 1.7e-9,
             'Tensor.dperp0': 1.7e-10,
             'Tensor.dperp1': 1.7e-10}

    volume_selection = {'unweighted_threshold': 25e6,
                        'use_unweighted': True,
                        'use_weighted': True,
                        'min_bval': 0,
                        'max_bval': 1.5e9 + 0.1e9}


class TensorExVivo(Tensor):

    name = 'Tensor-ExVivo'
    ex_vivo_suitable = True
    in_vivo_suitable = False
    description = 'The standard Tensor model with ex vivo defaults.'
    inits = {'Tensor.d': 1e-9,
             'Tensor.dperp0': 0.6e-10,
             'Tensor.dperp1': 0.6e-10}
    volume_selection = None



#!/usr/bin/env python
# PYTHON_ARGCOMPLETE_OK
"""Merge a list of volume images on the 4th dimension. Writes the result to a single image.

Please note that by default this will sort the list of volume names based on a natural key sort. This is
the most convenient option in the case of globbing files. You can disable this behaviour
using the flag --no-sort.
"""
import argparse
import glob
import os
from argcomplete.completers import FilesCompleter
import textwrap

from mdt.utils import volume_merge
from mdt.shell_utils import BasicShellApplication, get_argparse_extension_checker

__author__ = 'Robbert Harms'
__date__ = "2015-08-18"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


class VolumeMerge(BasicShellApplication):

    def _get_arg_parser(self):
        description = textwrap.dedent(__doc__)
        description += self._get_citation_message()

        epilog = textwrap.dedent("""
            Examples of use:
                mdt-volume-merge merged.nii.gz *.nii.gz
                mdt-volume-merge --no-sort merged.nii.gz *.nii.gz
        """)

        parser = argparse.ArgumentParser(description=description, epilog=epilog,
                                         formatter_class=argparse.RawTextHelpFormatter)

        parser.add_argument('-o', '--output_file', required=True,
                            action=get_argparse_extension_checker(['.nii', '.nii.gz', '.hdr', '.img']),
                            help='the filename of the output file').completer = \
            FilesCompleter(['nii', 'gz', 'hdr', 'img'], directories=False)

        parser.add_argument("input_files", metavar="input_files", nargs="+", type=str, help="The input images to merge")

        parser.add_argument('--sort', dest='sort', action='store_true',
                            help="Sort the input images using a natural sort (default)")
        parser.add_argument('--no-sort', dest='sort', action='store_false',
                            help="Do not sort the input images")
        parser.set_defaults(sort=True)

        parser.add_argument('--no-merge-order-file', dest='no_merge_order_file', action='store_true',
                            help="Do not write the merge order file")

        return parser

    def run(self, args):
        output_file = os.path.realpath(args.output_file)

        if os.path.isfile(output_file):
            os.remove(output_file)

        file_names = []
        for file in args.input_files:
            file_names.extend(glob.glob(file))

        concatenated_names = volume_merge(file_names, output_file, sort=args.sort)

        if not args.no_merge_order_file:
            info_output_file = os.path.splitext(output_file)[0].replace('.nii', '') + '_merge_order.txt'

            if os.path.isfile(info_output_file):
                os.remove(info_output_file)

            with open(info_output_file, 'w') as f:
                f.write('Files merged in this order:\n')
                for name in concatenated_names:
                    f.write(name + '\n')

if __name__ == '__main__':
    VolumeMerge().start()



import itertools
import numpy as np
from matplotlib.gridspec import GridSpec
from mdt.visualization.dict_conversion import SimpleClassConversion, IntConversion, SimpleDictConversion

__author__ = 'Robbert Harms'
__date__ = "2016-09-02"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


class GridLayout(object):

    def __init__(self, spacings=None):
        super(GridLayout, self).__init__()
        self.spacings = spacings or {'left': 0.10, 'right': 0.86,
                                     'top': 0.97, 'bottom': 0.04,
                                     'wspace': 0.5, 'hspace': 0.2}

        if self.spacings['top'] < self.spacings['bottom']:
            raise ValueError('The top ({}) can not be smaller than the bottom ({}) in the spacings'.format(
                self.spacings['top'], self.spacings['bottom']))
        if self.spacings['left'] > self.spacings['right']:
            raise ValueError('Left ({}) can not be larger than right ({}) in the spacings'.format(
                self.spacings['left'], self.spacings['right']))

    @classmethod
    def get_conversion_info(cls):
        return SimpleClassConversion(cls, cls._get_attribute_conversions())

    @classmethod
    def _get_attribute_conversions(cls):
        return {'spacings': SimpleDictConversion(desired_type=float)}

    def get_gridspec(self, figure, nmr_plots):
        """Get the grid layout specifier for the given figure using the given number of plots.

        Args:
            figure (Figure): the figure to add the axis to
            nmr_plots (int): the total number of plots

        Returns:
            GridLayoutSpecifier: the specifier we can ask new subplot axis from
        """

    def __eq__(self, other):
        if not isinstance(other, GridLayout):
            return NotImplemented
        return isinstance(other, type(self)) and other.spacings == self.spacings

    def __ne__(self, other):
        return not self.__eq__(other)


class GridLayoutSpecifier(object):

    def __init__(self, gridspec, figure, positions=None):
        """Create a grid layout specifier using the given gridspec and the given figure.

        Args:
            gridspec (GridSpec): the gridspec to use
            figure (Figure): the figure to generate subplots for
            positions (:class:`list`): if given, a list with grid spec indices for every requested axis
        """
        self.gridspec = gridspec
        self.figure = figure
        self.positions = positions

    def get_axis(self, index):
        gridspec_ind = self.gridspec[index]
        if self.positions is not None:
            gridspec_ind = self.gridspec[self.positions[index]]
        return self.figure.add_subplot(gridspec_ind)


class AutoGridLayout(GridLayout):

    def get_gridspec(self, figure, nmr_plots):
        rows, cols = self._get_row_cols_square(nmr_plots)
        return GridLayoutSpecifier(GridSpec(rows, cols, **self.spacings), figure)

    def _get_row_cols_square(self, nmr_plots):
        defaults = ((1, 1), (1, 2), (1, 3), (2, 2), (2, 3), (2, 3), (2, 3))
        if nmr_plots < len(defaults):
            return defaults[nmr_plots - 1]
        else:
            cols = np.ceil(nmr_plots / 3.0)
            rows = np.ceil(float(nmr_plots) / cols)
            rows = int(rows)
            cols = int(cols)
        return rows, cols


class Rectangular(GridLayout):

    def __init__(self, rows=None, cols=None, spacings=None):
        super(Rectangular, self).__init__(spacings=spacings)
        self.rows = rows
        self.cols = cols

        if self.rows is not None:
            self.rows = int(self.rows)
            if self.rows < 1:
                raise ValueError('The number of rows ({}) can not be smaller than 1.'.format(self.rows))

        if self.cols is not None:
            self.cols = int(self.cols)
            if self.cols < 1:
                raise ValueError('The number of columns ({}) can not be smaller than 1.'.format(self.rows))

    @classmethod
    def _get_attribute_conversions(cls):
        conversions = super(Rectangular, cls)._get_attribute_conversions()
        conversions.update({'rows': IntConversion(), 'cols': IntConversion()})
        return conversions

    def get_gridspec(self, figure, nmr_plots):
        rows = self.rows
        cols = self.cols

        if rows is None and cols is None:
            return AutoGridLayout(spacings=self.spacings).get_gridspec(figure, nmr_plots)

        if rows is None:
            rows = int(np.ceil(nmr_plots / cols))
        if cols is None:
            cols = int(np.ceil(nmr_plots / rows))

        if rows * cols < nmr_plots:
            cols = int(np.ceil(nmr_plots / rows))

        return GridLayoutSpecifier(GridSpec(rows, cols, **self.spacings), figure)

    def __eq__(self, other):
        if not isinstance(other, GridLayout):
            return NotImplemented
        return isinstance(other, type(self)) and other.rows == self.rows and other.cols == self.cols \
               and other.spacings == self.spacings


class LowerTriangular(GridLayout):

    def __init__(self, spacings=None):
        super(LowerTriangular, self).__init__(spacings=spacings)

    def get_gridspec(self, figure, nmr_plots):
        size, positions = self._get_size_and_position(nmr_plots)
        return GridLayoutSpecifier(GridSpec(size, size, **self.spacings), figure, positions=positions)

    def _get_size_and_position(self, nmr_plots):
        size = self._get_lowest_triangle_length(nmr_plots)

        positions = []
        for x, y in itertools.product(range(size), range(size)):
            if x >= y:
                positions.append(x * size + y)

        return size, positions

    @staticmethod
    def _get_lowest_triangle_length(nmr_plots):
        for n in range(1, nmr_plots):
            if 0.5 * (n ** 2 + n) >= nmr_plots:
                return n
        return nmr_plots


class SingleColumn(GridLayout):

    def get_gridspec(self, figure, nmr_plots):
        return GridLayoutSpecifier(GridSpec(nmr_plots, 1, **self.spacings), figure)


class SingleRow(GridLayout):

    def get_gridspec(self, figure, nmr_plots):
        return GridLayoutSpecifier(GridSpec(1, nmr_plots, **self.spacings), figure)



import yaml
from PyQt5.QtCore import pyqtSlot
from PyQt5.QtWidgets import QWidget

from mdt.gui.maps_visualizer.actions import NewConfigAction
from mdt.gui.maps_visualizer.design.ui_TabTextual import Ui_TabTextual
from mdt.gui.utils import blocked_signals
from mdt.visualization.maps.base import DataInfo, MapPlotConfig

__author__ = 'Robbert Harms'
__date__ = "2016-09-03"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


class TabTextual(QWidget, Ui_TabTextual):

    def __init__(self, controller, parent=None):
        super(TabTextual, self).__init__(parent)
        self.setupUi(self)

        self._controller = controller
        self._controller.new_data.connect(self.set_new_data)
        self._controller.new_config.connect(self.set_new_config)

        self.textConfigEdit.new_config.connect(self._config_from_string)
        self._update_status_indication(True)

        self._flags = {'updating_config_from_string': False}

    @pyqtSlot(DataInfo)
    def set_new_data(self, data_info):
        pass

    @pyqtSlot(MapPlotConfig)
    def set_new_config(self, config):
        with blocked_signals(self.textConfigEdit):
            if not self._flags['updating_config_from_string']:
                self.textConfigEdit.setPlainText(config.to_yaml())
                self._update_status_indication(True)

    @pyqtSlot(str)
    def _config_from_string(self, text):
        self._flags['updating_config_from_string'] = True
        text = text.replace('\t', ' '*4)
        try:
            if text.strip() != '':
                new_config = MapPlotConfig.from_yaml(text)
                new_config.validate(self._controller.get_data())

                self._controller.apply_action(NewConfigAction(new_config))
                self._update_status_indication(True)

        except Exception as exc:
            self._update_status_indication(False, str(exc))
            pass
        finally:
            self._flags['updating_config_from_string'] = False

    def _update_status_indication(self, status, status_message=''):
        if status:
            self.textConfigEdit.setStyleSheet("border: 1px solid green")
        else:
            self.textConfigEdit.setStyleSheet("border: 1px solid red")
        self.correctness_label.setText(status_message)



#!/usr/bin/env python
# PYTHON_ARGCOMPLETE_OK
"""This script is meant to update your home folder with the latest MDT models."""
import argparse
import os
import textwrap
import mdt
from mdt.configuration import get_config_dir
from mdt.shell_utils import BasicShellApplication

__author__ = 'Robbert Harms'
__date__ = "2015-08-18"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


class InitUserSettings(BasicShellApplication):

    def _get_arg_parser(self):
        description = textwrap.dedent(__doc__ + """

            The location we will write to is: {}
        """.format(os.path.dirname(get_config_dir())))
        description += mdt.shell_utils.get_citation_message()

        parser = argparse.ArgumentParser(description=description, formatter_class=argparse.RawTextHelpFormatter)

        parser.add_argument('--pass-if-exists', dest='pass_if_exists', action='store_true',
                            help="do nothing if the config dir exists (default)")
        parser.add_argument('--always-overwrite', dest='pass_if_exists', action='store_false',
                            help="always overwrite the config directory with the default settings")
        parser.set_defaults(pass_if_exists=False)

        return parser

    def run(self, args):
        mdt.init_user_settings(pass_if_exists=args.pass_if_exists)


if __name__ == '__main__':
    InitUserSettings().start()



import os

import numpy as np
from matplotlib import pyplot as plt
from mpl_toolkits.axes_grid1 import make_axes_locatable

from mdt import get_slice_in_dimension
from mdt.visualization.maps.base import Clipping, Scale, Point
from mdt.visualization.utils import MyColourBarTickLocator

__author__ = 'Robbert Harms'
__date__ = "2016-09-02"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


class MapsVisualizer(object):

    def __init__(self, data_info, figure):
        self._data_info = data_info
        self._figure = figure

    def render(self, plot_config):
        """Render all the maps to the figure. This is for use in GUI embedded situations.

        Returns:
            list of AxisData: the list with the drawn axes and the accompanying data
        """
        renderer = Renderer(self._data_info, self._figure, plot_config)
        renderer.render()
        return renderer.image_axes

    def to_file(self, file_name, plot_config, **kwargs):
        """Renders the figures to the given filename."""
        Renderer(self._data_info, self._figure, plot_config).render()

        if not os.path.isdir(os.path.dirname(file_name)):
            os.makedirs(os.path.dirname(file_name))

        kwargs['dpi'] = kwargs.get('dpi') or 300
        self._figure.savefig(file_name, **kwargs)

    def show(self, plot_config, block=True, maximize=False, window_title=None):
        """Show the data contained in this visualizer using the specifics in this function call.

        Args:
            plot_config (mdt.visualization.maps.base.MapPlotConfig): the plot configuration
            block (boolean): If we want to block after calling the plots or not. Set this to False if you
                do not want the routine to block after drawing. In doing so you manually need to block.
            maximize (boolean): if we want to display the window maximized or not
            window_title (str): the title of the window. If None, the default title is used
        """
        Renderer(self._data_info, self._figure, plot_config).render()

        if maximize:
            mng = plt.get_current_fig_manager()
            mng.window.showMaximized()

        if window_title:
            mng = plt.get_current_fig_manager()
            mng.canvas.set_window_title(window_title)

        if block:
            plt.show(True)


class AxisData(object):

    def __init__(self, axis, map_name, map_info, plot_config):
        """Contains a reference to a drawn matpotlib axis and to the accompanying data.

        Args:
            axis (Axis): the matpotlib axis
            map_name (str): the name/key of this map
            map_info (SingleMapInfo): the map information
            plot_config (MapPlotConfig): the map plot configuration
        """
        self.axis = axis
        self.map_name = map_name
        self._map_info = map_info
        self._plot_config = plot_config

    def coordinates_to_index(self, x, y):
        """Converts data coordinates to index coordinates of the array.

        Args:
            x (int): The x-coordinate in data coordinates.
            y (int): The y-coordinate in data coordinates.

        Returns
            x, y, z, v : Index coordinates of the map associated with the image.
        """
        shape = self._map_info.get_size_in_dimension(self._plot_config.dimension, self._plot_config.rotate)

        # correct for zoom
        x += self._plot_config.zoom.p0.x
        y += self._plot_config.zoom.p0.y

        # correct for flip upside down
        if not self._plot_config.flipud:
            y = self._map_info.get_max_y_index(self._plot_config.dimension, self._plot_config.rotate) - y

        # correct for displayed axis, the view is x-data on y-image and y-data on x-image
        x, y = y, x

        # rotate the point
        rotated = Point(x, y).rotate90((-1 * self._plot_config.rotate % 360) // 90)

        # translate the point back to a new origin
        if self._plot_config.rotate == 90:
            rotated.y = shape[1] + rotated.y
        elif self._plot_config.rotate == 180:
            rotated.x = shape[1] + rotated.x
            rotated.y = shape[0] + rotated.y
        elif self._plot_config.rotate == 270:
            rotated.x = shape[0] + rotated.x

        # create the index
        index = [rotated.x, rotated.y]
        index.insert(self._plot_config.dimension, self._plot_config.slice_index)

        if len(self._map_info.data.shape) > 3:
            if self._plot_config.volume_index < self._map_info.data.shape[3]:
                index.append(self._plot_config.volume_index)
            else:
                index.append(self._map_info.data.shape[3] - 1)

        return index

    def get_value(self, index):
        """Get the value of this axis data at the given index.

        Args:
            index (tuple)): the 3d or 4d index to the map corresponding to this axis data (x, y, z, [v])

        Returns:
            float: the value at the given index.
        """
        return self._map_info.data[tuple(index)]


class Renderer(object):

    def __init__(self, data_info, figure, plot_config):
        """Create a new renderer for the given volumes on the given figure using the given configuration.

        This renders the images with flipped upside down with the origin at the bottom left. The upside down flip
        is necessary to allow counter-clockwise rotation.

        Args:
            data_info (DataInfo): the information about the maps to show
            figure (Figure): the matplotlib figure to draw on
            plot_config (mdt.visualization.maps.base.MapPlotConfig): the plot configuration
        """
        self._data_info = data_info
        self._figure = figure
        self._plot_config = plot_config
        self.image_axes = []

    def render(self):
        """Render the maps"""
        grid_layout_specifier = self._plot_config.grid_layout.get_gridspec(
            self._figure, len(self._plot_config.maps_to_show))

        if self._plot_config.title:
            self._figure.suptitle(self._plot_config.title, fontsize=self._plot_config.font.size,
                                  family=self._plot_config.font.name)

        for ind, map_name in enumerate(self._plot_config.maps_to_show):
            axis = grid_layout_specifier.get_axis(ind)
            axis_data = self._render_map(map_name, axis)
            self.image_axes.append(axis_data)

    def _render_map(self, map_name, axis):
        axis.set_title(self._get_title(map_name), y=self._get_title_spacing(map_name))
        axis.axis('on' if self._plot_config.show_axis else 'off')

        data = self._get_image(map_name)
        if self._plot_config.rotate:
            data = np.rot90(data, self._plot_config.rotate // 90)

        if not self._plot_config.flipud:
            # by default we flipud to correct for matplotlib lower origin. If the user
            # sets flipud, we do not need to to it
            data = np.flipud(data)

        data = self._plot_config.zoom.apply(data)

        plot_options = self._get_map_plot_options(map_name)
        plot_options['origin'] = 'lower'
        plot_options['interpolation'] = self._plot_config.interpolation
        vf = axis.imshow(data, **plot_options)

        divider = make_axes_locatable(axis)
        colorbar_axis = divider.append_axes("right", size="5%", pad=0.05)

        self._add_colorbar(map_name, colorbar_axis, vf, self._get_map_attr(map_name, 'colorbar_label'))
        self._apply_font(axis, colorbar_axis)

        return AxisData(axis, map_name, self._data_info.map_info[map_name], self._plot_config)

    def _apply_font(self, image_axis, colorbar_axis):
        items = [image_axis.xaxis.label, image_axis.yaxis.label]
        items.extend(image_axis.get_xticklabels())
        items.extend(image_axis.get_yticklabels())
        items.extend(colorbar_axis.yaxis.get_ticklabels())

        for item in items:
            item.set_fontsize(self._plot_config.font.size - 2)
            item.set_family(self._plot_config.font.name)

        image_axis.title.set_fontsize(self._plot_config.font.size)
        image_axis.title.set_family(self._plot_config.font.name)

        colorbar_axis.yaxis.label.set_fontsize(self._plot_config.font.size)
        colorbar_axis.yaxis.label.set_family(self._plot_config.font.name)

        colorbar_axis.yaxis.offsetText.set_fontsize(self._plot_config.font.size - 3)
        colorbar_axis.yaxis.offsetText.set_family(self._plot_config.font.name)

    def _add_colorbar(self, map_name, axis, image_figure, colorbar_label):
        kwargs = dict(cax=axis, ticks=self._get_tick_locator(map_name))
        if colorbar_label:
            kwargs.update(dict(label=colorbar_label))

        cbar = plt.colorbar(image_figure, **kwargs)
        cbar.formatter.set_powerlimits((-3, 4))
        cbar.ax.yaxis.set_offset_position('left')
        cbar.update_ticks()

        if cbar.ax.get_yticklabels():
            cbar.ax.get_yticklabels()[-1].set_verticalalignment('top')
        return cbar

    def _get_map_attr(self, map_name, option, default=None):
        if map_name in self._plot_config.map_plot_options:
            value = getattr(self._plot_config.map_plot_options[map_name], option)
            if value:
                return value
        return default

    def _get_title(self, map_name):
        return self._get_map_attr(map_name, 'title', map_name)

    def _get_title_spacing(self, map_name):
        return 1 + self._get_map_attr(map_name, 'title_spacing', 0)

    def _get_map_plot_options(self, map_name):
        output_dict = {'vmin': self._data_info.maps[map_name].min(),
                       'vmax': self._data_info.maps[map_name].max(),
                       'cmap': self._get_map_attr(map_name, 'colormap', self._plot_config.colormap)}

        scale = self._get_map_attr(map_name, 'scale', Scale())
        if scale.use_max:
            output_dict['vmax'] = scale.vmax
        if scale.use_min:
            output_dict['vmin'] = scale.vmin

        return output_dict

    def _get_image(self, map_name):
        """Get the 2d image to display for the given data."""
        data = self._data_info.maps[map_name]

        dimension = self._plot_config.dimension
        slice_index = self._plot_config.slice_index
        volume_index = self._plot_config.volume_index

        def get_slice(data):
            slice = get_slice_in_dimension(data, dimension, slice_index)
            if len(slice.shape) > 2:
                if volume_index < slice.shape[2]:
                    slice = np.squeeze(slice[:, :, volume_index])
                else:
                    slice = np.squeeze(slice[:, :, slice.shape[2] - 1])
            return slice

        slice = get_slice(data)
        slice = self._get_map_attr(map_name, 'clipping', Clipping()).apply(slice)

        mask_name = self._get_map_attr(map_name, 'mask_name', self._plot_config.mask_name)
        if mask_name:
            slice = slice * (get_slice(self._data_info.maps[mask_name]) > 0)

        return slice

    def _get_tick_locator(self, map_name):
        min_val, max_val = self._data_info.maps[map_name].min(), self._data_info.maps[map_name].max()

        scale = self._get_map_attr(map_name, 'scale', Scale())
        if scale.use_max:
            max_val = scale.vmax
        if scale.use_min:
            min_val = scale.vmin

        return MyColourBarTickLocator(min_val, max_val, numticks=self._plot_config.colorbar_nmr_ticks)



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'generate_protocol_tab.ui'
#
# Created by: PyQt5 UI code generator 5.4.2
#
# WARNING! All changes made in this file will be lost!

from PyQt5 import QtCore, QtGui, QtWidgets

class Ui_GenerateProtocolTabContent(object):
    def setupUi(self, GenerateProtocolTabContent):
        GenerateProtocolTabContent.setObjectName("GenerateProtocolTabContent")
        GenerateProtocolTabContent.resize(827, 427)
        self.verticalLayout = QtWidgets.QVBoxLayout(GenerateProtocolTabContent)
        self.verticalLayout.setContentsMargins(-1, 11, -1, -1)
        self.verticalLayout.setObjectName("verticalLayout")
        self.verticalLayout_2 = QtWidgets.QVBoxLayout()
        self.verticalLayout_2.setSpacing(0)
        self.verticalLayout_2.setObjectName("verticalLayout_2")
        self.label = QtWidgets.QLabel(GenerateProtocolTabContent)
        font = QtGui.QFont()
        font.setPointSize(14)
        self.label.setFont(font)
        self.label.setObjectName("label")
        self.verticalLayout_2.addWidget(self.label)
        self.label_2 = QtWidgets.QLabel(GenerateProtocolTabContent)
        font = QtGui.QFont()
        font.setItalic(True)
        self.label_2.setFont(font)
        self.label_2.setObjectName("label_2")
        self.verticalLayout_2.addWidget(self.label_2)
        self.verticalLayout.addLayout(self.verticalLayout_2)
        self.line = QtWidgets.QFrame(GenerateProtocolTabContent)
        self.line.setFrameShadow(QtWidgets.QFrame.Sunken)
        self.line.setLineWidth(1)
        self.line.setFrameShape(QtWidgets.QFrame.HLine)
        self.line.setObjectName("line")
        self.verticalLayout.addWidget(self.line)
        self.protocol_table = QtWidgets.QTableWidget(GenerateProtocolTabContent)
        self.protocol_table.setToolTip("")
        self.protocol_table.setRowCount(0)
        self.protocol_table.setColumnCount(0)
        self.protocol_table.setObjectName("protocol_table")
        self.verticalLayout.addWidget(self.protocol_table)
        self.gridLayout = QtWidgets.QGridLayout()
        self.gridLayout.setContentsMargins(-1, 0, -1, -1)
        self.gridLayout.setObjectName("gridLayout")
        self.horizontalLayout_6 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_6.setContentsMargins(-1, 0, -1, -1)
        self.horizontalLayout_6.setObjectName("horizontalLayout_6")
        self.label_18 = QtWidgets.QLabel(GenerateProtocolTabContent)
        self.label_18.setObjectName("label_18")
        self.horizontalLayout_6.addWidget(self.label_18)
        self.nmrShells = QtWidgets.QLabel(GenerateProtocolTabContent)
        self.nmrShells.setObjectName("nmrShells")
        self.horizontalLayout_6.addWidget(self.nmrShells)
        spacerItem = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout_6.addItem(spacerItem)
        self.gridLayout.addLayout(self.horizontalLayout_6, 0, 3, 1, 1)
        self.horizontalLayout_7 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_7.setContentsMargins(-1, 0, -1, -1)
        self.horizontalLayout_7.setObjectName("horizontalLayout_7")
        self.label_13 = QtWidgets.QLabel(GenerateProtocolTabContent)
        self.label_13.setObjectName("label_13")
        self.horizontalLayout_7.addWidget(self.label_13)
        self.nmrColumns = QtWidgets.QLabel(GenerateProtocolTabContent)
        self.nmrColumns.setObjectName("nmrColumns")
        self.horizontalLayout_7.addWidget(self.nmrColumns)
        spacerItem1 = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout_7.addItem(spacerItem1)
        self.gridLayout.addLayout(self.horizontalLayout_7, 0, 4, 1, 1)
        self.horizontalLayout_3 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_3.setContentsMargins(-1, 0, -1, -1)
        self.horizontalLayout_3.setObjectName("horizontalLayout_3")
        self.label_8 = QtWidgets.QLabel(GenerateProtocolTabContent)
        self.label_8.setObjectName("label_8")
        self.horizontalLayout_3.addWidget(self.label_8)
        self.nmrRows = QtWidgets.QLabel(GenerateProtocolTabContent)
        self.nmrRows.setObjectName("nmrRows")
        self.horizontalLayout_3.addWidget(self.nmrRows)
        spacerItem2 = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout_3.addItem(spacerItem2)
        self.gridLayout.addLayout(self.horizontalLayout_3, 0, 0, 1, 1)
        self.horizontalLayout_5 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_5.setContentsMargins(-1, 0, -1, -1)
        self.horizontalLayout_5.setObjectName("horizontalLayout_5")
        self.label_17 = QtWidgets.QLabel(GenerateProtocolTabContent)
        self.label_17.setObjectName("label_17")
        self.horizontalLayout_5.addWidget(self.label_17)
        self.nmrWeighted = QtWidgets.QLabel(GenerateProtocolTabContent)
        self.nmrWeighted.setObjectName("nmrWeighted")
        self.horizontalLayout_5.addWidget(self.nmrWeighted)
        spacerItem3 = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout_5.addItem(spacerItem3)
        self.gridLayout.addLayout(self.horizontalLayout_5, 0, 2, 1, 1)
        self.horizontalLayout_4 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_4.setContentsMargins(-1, 0, -1, -1)
        self.horizontalLayout_4.setObjectName("horizontalLayout_4")
        self.label_16 = QtWidgets.QLabel(GenerateProtocolTabContent)
        self.label_16.setObjectName("label_16")
        self.horizontalLayout_4.addWidget(self.label_16)
        self.nmrUnweighted = QtWidgets.QLabel(GenerateProtocolTabContent)
        self.nmrUnweighted.setObjectName("nmrUnweighted")
        self.horizontalLayout_4.addWidget(self.nmrUnweighted)
        spacerItem4 = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout_4.addItem(spacerItem4)
        self.gridLayout.addLayout(self.horizontalLayout_4, 0, 1, 1, 1)
        self.horizontalLayout_9 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_9.setObjectName("horizontalLayout_9")
        self.label_4 = QtWidgets.QLabel(GenerateProtocolTabContent)
        self.label_4.setObjectName("label_4")
        self.horizontalLayout_9.addWidget(self.label_4)
        self.differentShells = QtWidgets.QLabel(GenerateProtocolTabContent)
        self.differentShells.setWordWrap(True)
        self.differentShells.setObjectName("differentShells")
        self.horizontalLayout_9.addWidget(self.differentShells)
        spacerItem5 = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout_9.addItem(spacerItem5)
        self.gridLayout.addLayout(self.horizontalLayout_9, 1, 0, 1, 5)
        self.verticalLayout.addLayout(self.gridLayout)
        self.line_3 = QtWidgets.QFrame(GenerateProtocolTabContent)
        self.line_3.setFrameShape(QtWidgets.QFrame.HLine)
        self.line_3.setFrameShadow(QtWidgets.QFrame.Sunken)
        self.line_3.setObjectName("line_3")
        self.verticalLayout.addWidget(self.line_3)
        self.horizontalLayout = QtWidgets.QHBoxLayout()
        self.horizontalLayout.setContentsMargins(-1, 0, -1, 0)
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.loadGB = QtWidgets.QPushButton(GenerateProtocolTabContent)
        self.loadGB.setObjectName("loadGB")
        self.horizontalLayout.addWidget(self.loadGB)
        self.loadProtocolButton = QtWidgets.QPushButton(GenerateProtocolTabContent)
        self.loadProtocolButton.setObjectName("loadProtocolButton")
        self.horizontalLayout.addWidget(self.loadProtocolButton)
        self.loadColumnButton = QtWidgets.QPushButton(GenerateProtocolTabContent)
        self.loadColumnButton.setObjectName("loadColumnButton")
        self.horizontalLayout.addWidget(self.loadColumnButton)
        self.saveButton = QtWidgets.QPushButton(GenerateProtocolTabContent)
        self.saveButton.setObjectName("saveButton")
        self.horizontalLayout.addWidget(self.saveButton)
        self.clearButton = QtWidgets.QPushButton(GenerateProtocolTabContent)
        self.clearButton.setObjectName("clearButton")
        self.horizontalLayout.addWidget(self.clearButton)
        spacerItem6 = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout.addItem(spacerItem6)
        self.verticalLayout.addLayout(self.horizontalLayout)

        self.retranslateUi(GenerateProtocolTabContent)
        QtCore.QMetaObject.connectSlotsByName(GenerateProtocolTabContent)

    def retranslateUi(self, GenerateProtocolTabContent):
        _translate = QtCore.QCoreApplication.translate
        GenerateProtocolTabContent.setWindowTitle(_translate("GenerateProtocolTabContent", "Form"))
        self.label.setText(_translate("GenerateProtocolTabContent", "Generate protocol file"))
        self.label_2.setText(_translate("GenerateProtocolTabContent", "Create a protocol file containing all your sequence information."))
        self.label_18.setText(_translate("GenerateProtocolTabContent", "# shells:"))
        self.nmrShells.setText(_translate("GenerateProtocolTabContent", "0"))
        self.label_13.setText(_translate("GenerateProtocolTabContent", "# columns:"))
        self.nmrColumns.setText(_translate("GenerateProtocolTabContent", "0"))
        self.label_8.setText(_translate("GenerateProtocolTabContent", "# rows:"))
        self.nmrRows.setText(_translate("GenerateProtocolTabContent", "0"))
        self.label_17.setText(_translate("GenerateProtocolTabContent", "# weighted:"))
        self.nmrWeighted.setText(_translate("GenerateProtocolTabContent", "0"))
        self.label_16.setText(_translate("GenerateProtocolTabContent", "# unweighted:"))
        self.nmrUnweighted.setText(_translate("GenerateProtocolTabContent", "0"))
        self.label_4.setText(_translate("GenerateProtocolTabContent", "Different shells:"))
        self.differentShells.setText(_translate("GenerateProtocolTabContent", "-"))
        self.loadGB.setText(_translate("GenerateProtocolTabContent", "Load g && b"))
        self.loadProtocolButton.setText(_translate("GenerateProtocolTabContent", "Load protocol"))
        self.loadColumnButton.setText(_translate("GenerateProtocolTabContent", "Add / Update column"))
        self.saveButton.setText(_translate("GenerateProtocolTabContent", "Save as"))
        self.clearButton.setText(_translate("GenerateProtocolTabContent", "Clear"))




from mdt.models.compartments import CompartmentConfig

__author__ = 'Francisco.Lagos'

# From protocol, if the signal is SE, we can setup TM = 0 in all the volumes,
# which returns to the standard SE signal decay


class ExpT1ExpT2STEAM(CompartmentConfig):
    """Generalised STEAM equation.

    From protocol, if the signal is SE, we can setup TM = 0 in all the volumes,
    which returns to the standard SE signal decay

    This equation can be used to calculate relaxation time (T1/T2) from spin echo (SE) and/or stimulated spin echo (STE)
    data. It is important to notice that in the protocol you have to define some parameters in a specific way:

    (1) For SE data, the original equation contains only the first refocusing pulse variable, but half of this value
        and in the power of two (sin(Refoc_fa1/2)**2). For that it is needed to define Refoc_fa2 = Refoc_fa1 and
        Refoc_fa1 has to be HALF of the used FA in the protocol (then, also Refoc_fa2). Also, the 0.5 factor is not included,
        then SEf (Spin echo flag) should be 0. Finally, TM (mixing time) has to be 0.
    (2) For STE data, this equation is used totally. Just SEf = 1.
    """

    parameter_list = ('SEf', 'TM', 'TE', 'flip_angle', 'Refoc_fa1', 'Refoc_fa2', 'T1', 'T2')
    cl_code = """
        return pow(0.5, SEf) * sin(flip_angle) * sin(Refoc_fa1) * sin(Refoc_fa2) * exp(-TE / T2) * exp(-TM / T1);
    """



from mdt.utils import ComplexNoiseStdEstimator, create_roi
from mdt.exceptions import NoiseStdEstimationNotPossible
import numpy as np

__author__ = 'Robbert Harms'
__date__ = "2016-04-11"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


class AverageOfAir_ExtendedMask(ComplexNoiseStdEstimator):

    def estimate(self, problem_data, **kwargs):
        """Calculate the standard deviation of the error using the air (voxels outside the brain),

        This procedure first finds the extreme points of the given brain mask in all dimensions. Next, it extends
        this mask (as a sort cross) in all dimensions to mask out the mask and possible ghostings. Finally we mask the
        first N voxels at the edges of the brain since the may be zero-filled. We use all the remaining voxels for
        the noise std calculation.

        We then calculate per voxel the std of the noise and use that to estimate the noise of the original complex
        image domain using:
            sigma_complex = sqrt(2.0 / (4.0 - PI)) * stddev(signal in background region)

        Finally, we take the median value of all calculated std's.
        """
        voxels = self._get_air_voxels(problem_data)

        if not len(voxels):
            raise NoiseStdEstimationNotPossible('No voxels in air found.')

        return np.median(np.sqrt(2.0 / (4.0 - np.pi)) * np.std(voxels, axis=1))

    def _get_air_voxels(self, problem_data, border_offset=3):
        """Get a two dimensional list with all the voxels in the air.

        Returns:
            ndarray: The first dimension is the list of voxels, the second the signal per voxel.
        """
        indices = np.where(problem_data.mask > 0)
        max_dims = np.max(indices, axis=1)
        min_dims = np.min(indices, axis=1)

        mask = np.copy(problem_data.mask)

        mask[min_dims[0]:max_dims[0]] = True
        mask[:, min_dims[1]:max_dims[1], :] = True
        mask[..., min_dims[2]:max_dims[2]] = True

        mask[0:border_offset] = True
        mask[-border_offset:] = True
        mask[:, 0:border_offset, :] = True
        mask[:, -border_offset:, :] = True
        mask[..., 0:border_offset] = True
        mask[..., -border_offset:] = True

        return create_roi(problem_data.dwi_volume, np.invert(mask))

    def __str__(self):
        return __name__



import logging
from copy import deepcopy

import numpy as np
import six

from mdt.components_loader import ComponentConfig, ComponentBuilder, method_binding_meta
from mdt.model_protocol_problem import MissingColumns, InsufficientShells
from mdt.models.base import DMRIOptimizable
from mdt.models.parsers.CompositeModelExpressionParser import parse
from mdt.protocols import VirtualColumnB
from mdt.utils import create_roi, calculate_information_criterions
from mot.cl_data_type import CLDataType
from mot.model_building.cl_functions.model_functions import Weight
from mot.cl_routines.mapping.loglikelihood_calculator import LogLikelihoodCalculator
from mot.model_building.data_adapter import SimpleDataAdapter
from mot.model_building.evaluation_models import OffsetGaussianEvaluationModel
from mot.model_building.model_builders import SampleModelBuilder
from mot.model_building.parameter_functions.dependencies import WeightSumToOneRule, SimpleAssignment
from mot.model_building.trees import CompartmentModelTree

__author__ = 'Robbert Harms'
__date__ = "2014-10-26"
__license__ = "LGPL v3"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


class DMRICompositeModel(SampleModelBuilder, DMRIOptimizable):

    def __init__(self, model_name, model_tree, evaluation_model, signal_noise_model=None, problem_data=None,
                 add_default_weights_dependency=True):
        """Create a composite dMRI sample model.

        This also implements the perturbation interface to allow perturbation of the data during meta-optimization.

        It furthermore implements some protocol check functions. These are used by the fit_model functions in MDT
        to check if the protocol is correct for the model we try to fit.

        Args:
            add_default_weights_dependency (boolean): if we want to add the default weights dependency to this model
                or not, by default we use this.

        Attributes:
            required_nmr_shells (int): Define the minimum number of unique shells necessary for this model.
                The default is false, which means that we don't check for this.
        """
        self._add_default_weights_dependency = add_default_weights_dependency
        super(DMRICompositeModel, self).__init__(model_name, model_tree, evaluation_model, signal_noise_model,
                                                 problem_data=problem_data)
        self.required_nmr_shells = False
        self._logger = logging.getLogger(__name__)
        self._original_problem_data = None

    def set_problem_data(self, problem_data):
        """Overwrites the super implementation by adding a call to _prepare_problem_data() before the problem data is
        added to the model.
        """
        self._check_data_consistency(problem_data)
        self._original_problem_data = problem_data
        return super(DMRICompositeModel, self).set_problem_data(self._prepare_problem_data(problem_data))

    def _get_variable_data(self):
        var_data_dict = super(DMRICompositeModel, self)._get_variable_data()

        if self._problem_data.gradient_deviations is not None:
            if len(self._problem_data.gradient_deviations.shape) > 2:
                grad_dev = create_roi(self._problem_data.gradient_deviations, self._problem_data.mask)
            else:
                grad_dev = np.copy(self._problem_data.gradient_deviations)

            self._logger.info('Using the gradient deviations in the model optimization.')

            # adds the eye(3) matrix to every grad dev, so we don't have to do it in the kernel.
            # Flattening an eye(3) matrix gives the same result with F and C ordering, I nevertheless put the ordering
            # here to emphasize that the gradient deviations matrix is in Fortran (column-major) order.
            grad_dev += np.eye(3).flatten(order='F')

            if self.problems_to_analyze is not None:
                grad_dev = grad_dev[self.problems_to_analyze, ...]

            adapter = SimpleDataAdapter(grad_dev, CLDataType.from_string('mot_float_type*'), self._get_mot_float_type())
            var_data_dict.update({'gradient_deviations': adapter})

        return var_data_dict

    def is_protocol_sufficient(self, protocol=None):
        """See ProtocolCheckInterface"""
        return not self.get_protocol_problems(protocol=protocol)

    def get_protocol_problems(self, protocol=None):
        """See ProtocolCheckInterface"""
        if protocol is None:
            protocol = self._problem_data.protocol

        problems = []

        missing_columns = [name for name in self.get_required_protocol_names() if not protocol.has_column(name)]
        if missing_columns:
            problems.append(MissingColumns(missing_columns))

        try:
            shells = protocol.get_nmr_shells()
            if shells < self.required_nmr_shells:
                problems.append(InsufficientShells(self.required_nmr_shells, shells))
        except KeyError:
            pass

        return problems

    def _set_default_dependencies(self):
        super(DMRICompositeModel, self)._set_default_dependencies()

        if self._add_default_weights_dependency:
            names = [w.name + '.w' for w in self._get_weight_models()]
            if len(names):
                self.add_parameter_dependency(names[0], WeightSumToOneRule(names[1:]))

    def _get_weight_models(self):
        return [cm for cm in self._model_tree.get_compartment_models() if isinstance(cm, Weight)]

    def _get_pre_model_expression_eval_code(self):
        if self._can_use_gradient_deviations():
            s = '''
                mot_float_type4 _new_gradient_vector_raw = _get_new_gradient_raw(g, data->var_data_gradient_deviations);
                mot_float_type _new_gradient_vector_length = length(_new_gradient_vector_raw);
                g = _new_gradient_vector_raw/_new_gradient_vector_length;
            '''

            if 'b' in list(self._get_protocol_data().keys()):
                s += 'b *= _new_gradient_vector_length * _new_gradient_vector_length;' + "\n"

            if 'G' in list(self._get_protocol_data().keys()):
                s += 'G *= _new_gradient_vector_length;' + "\n"

            if 'q' in list(self._get_protocol_data().keys()):
                s += 'q *= _new_gradient_vector_length;' + "\n"

            return s

    def _get_pre_model_expression_eval_function(self):
        if self._can_use_gradient_deviations():
            return '''
                #ifndef GET_NEW_GRADIENT_RAW
                #define GET_NEW_GRADIENT_RAW
                mot_float_type4 _get_new_gradient_raw(mot_float_type4 g,
                                                   global const mot_float_type* const gradient_deviations){

                    const mot_float_type4 il_0 = (mot_float_type4)(gradient_deviations[0], gradient_deviations[3],
                                                             gradient_deviations[6], 0.0);

                    const mot_float_type4 il_1 = (mot_float_type4)(gradient_deviations[1], gradient_deviations[4],
                                                             gradient_deviations[7], 0.0);

                    const mot_float_type4 il_2 = (mot_float_type4)(gradient_deviations[2], gradient_deviations[5],
                                                             gradient_deviations[8], 0.0);

                    return (mot_float_type4)(dot(il_0, g), dot(il_1, g), dot(il_2, g), 0.0);
                }
                #endif //GET_NEW_GRADIENT_RAW
            '''

    def _can_use_gradient_deviations(self):
        return self._problem_data.gradient_deviations is not None \
               and 'g' in list(self._get_protocol_data().keys())

    def _add_finalizing_result_maps(self, results_dict):
        super(DMRICompositeModel, self)._add_finalizing_result_maps(results_dict)

        log_likelihood_calc = LogLikelihoodCalculator()
        log_likelihoods = log_likelihood_calc.calculate(self, results_dict)

        k = self.get_nmr_estimable_parameters()
        n = self._problem_data.get_nmr_inst_per_problem()

        results_dict.update({'LogLikelihood': log_likelihoods})
        results_dict.update(calculate_information_criterions(log_likelihoods, k, n))

    def _prepare_problem_data(self, problem_data):
        """Update the problem data to make it suitable for this model.

        Some of the models in diffusion MRI can only handle a subset of all volumes. For example, the S0 model
        can only work with the unweigthed signals, or the Tensor model that can only handle a b-value up to 1.5e9 s/m^2.

        Overwrite this function to limit the problem data to a suitable range.

        Args:
            problem_data (DMRIProblemData): the problem data set by the user

        Returns:
            DMRIProblemData: either the same problem data or a changed copy.
        """
        protocol = problem_data.protocol
        indices = self._get_suitable_volume_indices(problem_data)

        if len(indices) != protocol.length:
            self._logger.info('For this model, {}, we will use a subset of the protocol and DWI.'.format(self._name))
            self._logger.info('Using {} out of {} volumes, indices: {}'.format(
                len(indices), protocol.length, str(indices).replace('\n', '').replace('[  ', '[')))

            new_protocol = protocol.get_new_protocol_with_indices(indices)
            new_dwi_volume = problem_data.dwi_volume[..., indices]
            return problem_data.copy_with_updates(new_protocol, new_dwi_volume)
        else:
            self._logger.info('No model protocol options to apply, using original protocol.')
        return problem_data

    def _check_data_consistency(self, problem_data):
        """Check the problem data for any strange anomalies.

        We do this here so that implementing models can add additional consistency checks, or skip the checks.
        Also, by doing this here instead of in the Protocol class we ensure that the warnings end up in the log file.
        The final argument for putting this here is that I do not want any log output in the protocol tab.

        Args:
            problem_data (DMRIProblemData): the problem data to analyze.
        """
        protocol = problem_data.protocol

        def warn(warning):
            self._logger.warning('{}, proceeding with seemingly inconsistent values.'.format(warning))

        if 'TE' in protocol and 'TR' in protocol:
            if any(np.greater(protocol['TE'], protocol['TR'])):
                warn('Volumes detected where TE > TR')

        if 'TE' in protocol:
            if any(np.greater_equal(protocol['TE'], 1)):
                warn('Volumes detected where TE >= 1 second')

        if 'TR' in protocol:
            if any(np.greater_equal(protocol['TR'], 50)):
                warn('Volumes detected where TR >= 50 seconds')

        if 'delta' in protocol and 'Delta' in protocol and any(map(protocol.is_column_real, ['delta', 'Delta'])):
            if any(np.greater_equal(protocol['delta'], protocol['Delta'])):
                warn('Volumes detected where (small) delta >= (big) Delta')

        if 'Delta' in protocol and 'TE' in protocol:
            if any(np.greater_equal(protocol['Delta'], protocol['TE'])):
                warn('Volumes detected where (big) Delta >= TE')

        if all(map(protocol.is_column_real, ['G', 'delta', 'Delta', 'b'])):
            if not np.allclose(VirtualColumnB().get_values(protocol), protocol['b']):
                warn('Estimated b-values (from G, Delta, delta) differ from given b-values')

    def _get_suitable_volume_indices(self, problem_data):
        """Usable in combination with _prepare_problem_data, return the suitable volume indices.

        Get a list of volume indices that the model can use. This function is meant to remove common boilerplate code
        from writing your own _prepare_problem_data object.

        Args:
            problem_data (DMRIProblemData): the problem data set by the user

        Returns:
            list: the list of indices we want to use for this model.
        """
        return list(range(problem_data.protocol.length))


class DMRICompositeModelConfig(ComponentConfig):
    """The cascade config to inherit from.

    These configs are loaded on the fly by the DMRICompositeModelBuilder

    Attributes:
        name (str): the name of the model, defaults to the class name
        in_vivo_suitable (boolean): flag indicating if the model is suitable for in vivo data
        ex_vivo_suitable (boolean): flag indicating if the model is suitable for ex vivo data
        description (str): model description
        post_optimization_modifiers (list): a list of modification callbacks for use after optimization. Example:

            .. code-block:: python

                post_optimization_modifiers = [('SNIF', lambda d: 1 - d['Wcsf.w']),
                                           ...]
        dependencies (list): the dependencies between model parameters. Example:

            .. code-block:: python

                dependencies = [('Noddi_EC.kappa', SimpleAssignment('Noddi_IC.kappa')),
                                ('NODDI_EC.theta', 'NODDI_IC.theta')
                                ...]

            If a string is given this is interpreted as a SimpleAssignment dependency.
            In the example shown here both the kappa and theta parameters are dependend in the same way.

        model_expression (str): the model expression. For the syntax see:
            mdt.models.parsers.CompositeModelExpression.ebnf
        evaluation_model (EvaluationModel): the evaluation model to use during optimization
        signal_noise_model (SignalNoiseModel): optional signal noise decorator
        inits (dict): indicating the initialization values for the parameters. Example:

            .. code-block:: python

                inits = {'Stick.theta: pi}

        fixes (dict): indicating the constant value for the given parameters. Example:

            .. code-block:: python

                fixes = {'Ball.d': 3.0e-9}

        upper_bounds (dict): indicating the upper bounds for the given parameters. Example:

            .. code-block:: python

                upper_bounds = {'Stick.theta': pi}

        lower_bounds (dict): indicating the lower bounds for the given parameters. Example:

            .. code-block:: python

                lower_bounds = {'Stick.theta': 0}

        parameter_transforms (dict): the parameter transform to use for a specific parameter. Can also be
            a python callback function accepting as single parameter 'self', a reference to the build model.
            This overwrites the default parameter transform of the specified parameter to the given transformation.
            Example:

            .. code-block:: python

                parameter_transforms = {
                    'Tensor.dperp0': SinSqrClampTransform(),
                    'Tensor.dperp1': lambda self: SinSqrClampDependentTransform(
                                                    [(self, self._get_parameter_by_name('Tensor.dperp0'))])
                }

        add_default_weights_dependency (boolean): set to False to disable the automatic Weight-sum-to-one dependency.
            By default it is True and we add them.

        volume_selection (dict): the volume selection by this model. This can be used to limit the volumes used
            in the analysis to only the volumes included in the specification.
            Set to None, or an empty dict to disable.
            The options available are:

               * ``unweighted_threshold`` (float): the threshold differentiating between
                 weighted and unweighted volumes
               * ``use_unweighted`` (bool): if we want to use unweighted volumes or not
               * ``use_weighted`` (bool): if we want to use the diffusion weigthed volumes or not
               * ``min_bval`` (float): the minimum b-value to include
               * ``max_bval`` (float): the maximum b-value to include

            If the method ``_get_suitable_volume_indices`` is overwritten, this does nothing.
    """
    name = ''
    in_vivo_suitable = True
    ex_vivo_suitable = True
    description = ''
    post_optimization_modifiers = []
    dependencies = []
    model_expression = ''
    evaluation_model = OffsetGaussianEvaluationModel()
    signal_noise_model = None
    inits = {}
    fixes = {}
    upper_bounds = {}
    lower_bounds = {}
    parameter_transforms = {}
    add_default_weights_dependency = True
    volume_selection = None

    @classmethod
    def meta_info(cls):
        meta_info = deepcopy(ComponentConfig.meta_info())
        meta_info.update({'name': cls.name,
                          'in_vivo_suitable': cls.in_vivo_suitable,
                          'ex_vivo_suitable': cls.ex_vivo_suitable,
                          'description': cls.description})
        return meta_info


class DMRICompositeModelBuilder(ComponentBuilder):

    def create_class(self, template):
        """Creates classes with as base class DMRICompositeModel

        Args:
            template (DMRICompositeModelConfig): the composite model config template
                to use for creating the class with the right init settings.
        """
        class AutoCreatedDMRICompositeModel(method_binding_meta(template, DMRICompositeModel)):

            def __init__(self, *args):
                super(AutoCreatedDMRICompositeModel, self).__init__(
                    deepcopy(template.name),
                    CompartmentModelTree(parse(template.model_expression)),
                    deepcopy(template.evaluation_model),
                    signal_noise_model=deepcopy(template.signal_noise_model),
                    add_default_weights_dependency=template.add_default_weights_dependency)

                self.add_parameter_dependencies(_resolve_dependencies(deepcopy(template.dependencies)))
                self.add_post_optimization_modifiers(deepcopy(template.post_optimization_modifiers))

                for full_param_name, value in template.inits.items():
                    self.init(full_param_name, deepcopy(value))

                for full_param_name, value in template.fixes.items():
                    self.fix(full_param_name, deepcopy(value))

                for full_param_name, value in template.lower_bounds.items():
                    self.set_lower_bound(full_param_name, deepcopy(value))

                for full_param_name, value in template.upper_bounds.items():
                    self.set_upper_bound(full_param_name, deepcopy(value))

                for full_param_name, value in template.parameter_transforms.items():
                    if hasattr(value, '__call__'):
                        self.set_parameter_transform(full_param_name, value(self))
                    else:
                        self.set_parameter_transform(full_param_name, deepcopy(value))

            def _get_suitable_volume_indices(self, problem_data):
                volume_selection = template.volume_selection

                if not volume_selection:
                    return super(AutoCreatedDMRICompositeModel, self)._get_suitable_volume_indices(problem_data)

                use_unweighted = volume_selection.get('use_unweighted', True)
                use_weighted = volume_selection.get('use_weighted', True)
                unweighted_threshold = volume_selection.get('unweighted_threshold', 25e6)

                protocol = problem_data.protocol

                if protocol.has_column('g') and protocol.has_column('b'):
                    if use_weighted:
                        if 'min_bval' in volume_selection and 'max_bval' in volume_selection:
                            protocol_indices = protocol.get_indices_bval_in_range(start=volume_selection['min_bval'],
                                                                                  end=volume_selection['max_bval'])
                        else:
                            protocol_indices = protocol.get_weighted_indices(unweighted_threshold)
                    else:
                        protocol_indices = []

                    if use_unweighted:
                        protocol_indices = list(protocol_indices) + \
                                           list(protocol.get_unweighted_indices(unweighted_threshold))
                else:
                    return list(range(protocol.length))

                return np.unique(protocol_indices)

        return AutoCreatedDMRICompositeModel


def _resolve_dependencies(dependencies):
    """Resolve string dependencies to SimpleAssignment objects in the list of dependencies.

    Args:
        dependencies (list): the dependencies to resolve strings in

    Returns:
        list: the list of dependencies with dependency a proper object
    """
    return_list = []
    for param, dependency in dependencies:
        if isinstance(dependency, six.string_types):
            return_list.append((param, SimpleAssignment(dependency)))
        else:
            return_list.append((param, dependency))
    return return_list



from mdt.models.composite import DMRICompositeModelConfig
from mot.model_building.parameter_functions.dependencies import AbstractParameterDependency

__author__ = 'Robbert Harms'
__date__ = "2015-06-22"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


class NODDITortuosityParameterDependency(AbstractParameterDependency):

    def __init__(self, d, w_ec, w_ic, ):
        self._d = d
        self._w_ec = w_ec
        self._w_ic = w_ic

    @property
    def pre_transform_code(self):
        return '''
            mot_float_type _tortuosity_mult_{d} = {w_ec} / ({w_ec} + {w_ic});
            if(!isnormal(_tortuosity_mult_{d})){{
                _tortuosity_mult_{d} = 0.01;
            }}
        '''.format(d=self._d, w_ec=self._w_ec, w_ic=self._w_ic)

    @property
    def assignment_code(self):
        return '{d} * _tortuosity_mult_{d}'.format(d=self._d)

    @property
    def fixed(self):
        return True

    @property
    def has_side_effects(self):
        return False


class NODDI(DMRICompositeModelConfig):

    ex_vivo_suitable = False
    description = 'The standard NODDI model'

    model_expression = '''
        S0 * ((Weight(w_csf) * Ball) +
              (Weight(w_ic) * NODDI_IC) +
              (Weight(w_ec) * NODDI_EC)
              )
    '''

    fixes = {'NODDI_IC.d': 1.7e-9,
             'NODDI_IC.R': 0.0,
             'NODDI_EC.d': 1.7e-9,
             'Ball.d': 3.0e-9}

    dependencies = (
        ('NODDI_EC.dperp0', NODDITortuosityParameterDependency('NODDI_EC.d', 'w_ec.w', 'w_ic.w')),
        ('NODDI_EC.kappa', 'NODDI_IC.kappa'),
        ('NODDI_EC.theta', 'NODDI_IC.theta'),
        ('NODDI_EC.phi', 'NODDI_IC.phi')
    )

    post_optimization_modifiers = [
        ('NDI', lambda d: d['w_ic.w'] / (d['w_ic.w'] + d['w_ec.w'])),
        ('ODI', lambda d: d['NODDI_IC.odi'])
    ]

